import datetime as dt
from collections import defaultdict
from itertools import chain

from django.db.models import Q, Sum
from django.db.models.functions import ExtractMonth, ExtractYear

from workbench.awt.reporting import full_time_equivalents_by_month
from workbench.invoices.models import Invoice
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.offers.models import Offer
from workbench.projects.models import Project, Service
from workbench.reporting.models import Accruals
from workbench.tools.formats import Z1, Z2


def gross_profit_by_month(date_range):
    profit = defaultdict(lambda: Z2)
    for result in (
        Invoice.objects.invoiced()
        .order_by()
        .filter(invoiced_on__range=date_range)
        .annotate(year=ExtractYear("invoiced_on"), month=ExtractMonth("invoiced_on"))
        .values("year", "month")
        .annotate(Sum("total_excl_tax"))
    ):
        profit[(result["year"], result["month"])] += result["total_excl_tax__sum"]
    return profit


def third_party_costs_by_month(date_range):
    costs = defaultdict(lambda: Z2)

    for result in (
        LoggedCost.objects.filter(
            rendered_on__range=date_range,
            third_party_costs__isnull=False,
            invoice_service__isnull=True,
        )
        .order_by()
        .annotate(year=ExtractYear("rendered_on"), month=ExtractMonth("rendered_on"))
        .values("year", "month")
        .annotate(Sum("third_party_costs"))
    ):
        costs[(result["year"], result["month"])] -= result["third_party_costs__sum"]

    for result in (
        Invoice.objects.invoiced()
        .filter(~Q(type=Invoice.DOWN_PAYMENT))
        .order_by()
        .filter(Q(invoiced_on__range=date_range), ~Q(third_party_costs=Z2))
        .annotate(year=ExtractYear("invoiced_on"), month=ExtractMonth("invoiced_on"))
        .values("year", "month")
        .annotate(Sum("third_party_costs"))
    ):
        costs[(result["year"], result["month"])] -= result["third_party_costs__sum"]

    return costs


def accruals_by_month(date_range):
    accruals = {(0, 0): {"accrual": Z2, "delta": None}}
    for accrual in Accruals.objects.order_by("cutoff_date"):
        accruals[(accrual.cutoff_date.year, accrual.cutoff_date.month)] = {
            "accrual": accrual.accruals,
            "delta": None,
        }

    today = dt.date.today()
    accruals[(today.year, today.month)] = {
        "accrual": Accruals.objects.accruals(cutoff_date=today),
        "delta": None,
    }

    dates = list(sorted(accruals))
    for this, next in zip(dates, dates[1:]):
        accruals[next]["delta"] = accruals[next]["accrual"] - accruals[this]["accrual"]

    return {
        month: accrual
        for month, accrual in accruals.items()
        if month != (0, 0)
        and date_range[0] <= dt.date(month[0], month[1], 1) <= date_range[1]
    }


def gross_margin_by_month(date_range):
    gross = gross_profit_by_month(date_range)
    third = third_party_costs_by_month(date_range)
    accruals = accruals_by_month(date_range)
    fte = full_time_equivalents_by_month()

    months = sorted(set(chain.from_iterable([gross, third, accruals])))
    profit = []
    for month in months:
        date = dt.date(month[0], month[1], 1)
        row = {
            "month": month,
            "key": "%s-%s" % month,
            "date": date,
            "gross_profit": gross[month],
            "third_party_costs": third[month],
            "accruals": accruals.get(month) or {"accrual": None, "delta": Z2},
            "fte": fte.get(date, Z2),
        }
        row["gross_margin"] = (
            row["gross_profit"] + row["third_party_costs"] + row["accruals"]["delta"]
        )
        row["margin_per_fte"] = row["gross_margin"] / row["fte"] if row["fte"] else None
        profit.append(row)

    return profit


def service_hours_in_open_orders():
    return (
        Service.objects.budgeted()
        .filter(project__type=Project.ORDER, project__closed_on__isnull=True)
        .order_by()
        .aggregate(h=Sum("service_hours"))["h"]
        or Z1
    )


def logged_hours_in_open_orders():
    return (
        LoggedHours.objects.filter(
            service__project__type=Project.ORDER,
            service__project__closed_on__isnull=True,
        )
        .order_by()
        .aggregate(h=Sum("hours"))["h"]
        or Z1
    )


def sent_invoices_total():
    return (
        Invoice.objects.filter(status=Invoice.SENT)
        .order_by()
        .aggregate(t=Sum("total_excl_tax"))["t"]
        or Z2
    )


def open_offers_total():
    return (
        Offer.objects.offered().order_by().aggregate(t=Sum("total_excl_tax"))["t"] or Z2
    )



# Generated by Django 2.1.7 on 2019-03-06 20:04

import datetime
from decimal import Decimal

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contacts", "0003_auto_20190305_0809"),
        ("invoices", "0003_auto_20190304_2247"),
    ]

    operations = [
        migrations.CreateModel(
            name="RecurringInvoice",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subtotal",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="subtotal",
                    ),
                ),
                (
                    "discount",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="discount",
                    ),
                ),
                (
                    "liable_to_vat",
                    models.BooleanField(
                        default=True,
                        help_text="For example invoices to foreign institutions are not liable to VAT.",
                        verbose_name="liable to VAT",
                    ),
                ),
                (
                    "tax_rate",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("7.7"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="tax rate",
                    ),
                ),
                (
                    "total",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="total",
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("postal_address", models.TextField(verbose_name="postal address")),
                (
                    "starts_on",
                    models.DateField(
                        default=datetime.date.today, verbose_name="starts on"
                    ),
                ),
                (
                    "ends_on",
                    models.DateField(blank=True, null=True, verbose_name="ends on"),
                ),
                (
                    "periodicity",
                    models.CharField(
                        choices=[
                            ("yearly", "yearly"),
                            ("quarterly", "quarterly"),
                            ("monthly", "monthly"),
                            ("weekly", "weekly"),
                        ],
                        max_length=20,
                        verbose_name="periodicity",
                    ),
                ),
                (
                    "next_period_starts_on",
                    models.DateField(
                        blank=True, null=True, verbose_name="next period starts on"
                    ),
                ),
                (
                    "contact",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contacts.Person",
                        verbose_name="contact",
                    ),
                ),
                (
                    "customer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contacts.Organization",
                        verbose_name="customer",
                    ),
                ),
                (
                    "owned_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="responsible",
                    ),
                ),
            ],
            options={
                "verbose_name": "recurring invoice",
                "verbose_name_plural": "recurring invoices",
                "ordering": ["customer__name", "title"],
            },
        ),
        migrations.RunSQL("SELECT audit_audit_table('invoices_recurringinvoice');", ""),
    ]



from collections import defaultdict

from django.db import connections
from django.db.models import Sum

from workbench.accounts.models import User
from workbench.logbook.models import LoggedHours
from workbench.offers.models import Offer
from workbench.projects.models import Project
from workbench.tools.formats import Z0, Z1


def green_hours(date_range, *, users=None):
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """\
WITH
service AS (
  SELECT ps.project_id, SUM(service_hours) AS hours
  FROM projects_service ps
  LEFT OUTER JOIN offers_offer o ON ps.offer_id=o.id
  WHERE ps.offer_id IS NULL OR o.status != %s
  GROUP BY ps.project_id
),
logged AS (
  SELECT project_id, SUM(hours) AS hours FROM logbook_loggedhours lh
  LEFT JOIN projects_service ps ON lh.service_id=ps.id
  GROUP BY project_id
)
SELECT logged.project_id, service.hours / logged.hours
FROM logged
LEFT JOIN service ON logged.project_id=service.project_id
WHERE service.hours / logged.hours < 1;
            """,
            [Offer.DECLINED],
        )
        green_hours_factor = defaultdict(lambda: Z0 + 1, cursor)

    within = defaultdict(lambda: defaultdict(lambda: Z1))
    project_ids = set()
    user_ids = set()

    for row in (
        LoggedHours.objects.filter(rendered_on__range=date_range)
        .order_by()
        .values("service__project", "rendered_by")
        .annotate(Sum("hours"))
    ):
        project_ids.add(row["service__project"])
        user_ids.add(row["rendered_by"])
        within[row["service__project"]][row["rendered_by"]] = row["hours__sum"]

    green = defaultdict(lambda: Z1)
    red = defaultdict(lambda: Z1)
    maintenance = defaultdict(lambda: Z1)
    internal = defaultdict(lambda: Z1)

    for project_id, type in Project.objects.filter(id__in=project_ids).values_list(
        "id", "type"
    ):
        for user_id, hours in within[project_id].items():
            if type == Project.INTERNAL:
                internal[user_id] += hours
            elif type == Project.MAINTENANCE:
                maintenance[user_id] += hours
            else:
                green[user_id] += green_hours_factor[project_id] * hours
                red[user_id] += (1 - green_hours_factor[project_id]) * hours

    def data(user_id):
        ret = {
            "green": green[user_id],
            "red": red[user_id],
            "maintenance": maintenance[user_id],
            "internal": internal[user_id],
        }
        ret["total"] = sum(ret.values())
        ret["percentage"] = (
            100 * (ret["green"] + ret["maintenance"]) / ret["total"]
            if ret["total"]
            else 0
        )
        return ret

    users = users or User.objects.filter(id__in=user_ids)
    ret = {}
    for user in users:
        ret[user] = data(user.id)
        green[0] += green[user.id]
        red[0] += red[user.id]
        maintenance[0] += maintenance[user.id]
        internal[0] += internal[user.id]
    ret[0] = data(0)
    return sorted((user, rec) for user, rec in ret.items() if rec["total"])


def green_hours_by_month():
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """\
DROP TABLE IF EXISTS green_hours_factor;
CREATE TEMPORARY TABLE green_hours_factor AS
WITH
service AS (
  SELECT ps.project_id, SUM(service_hours) AS hours
  FROM projects_service ps
  LEFT OUTER JOIN offers_offer o ON ps.offer_id=o.id
  WHERE ps.offer_id IS NULL OR o.status!=40 -- Declined
  GROUP BY ps.project_id
),
logged AS (
  SELECT project_id, SUM(hours) AS hours FROM logbook_loggedhours lh
  LEFT JOIN projects_service ps ON lh.service_id=ps.id
  GROUP BY project_id
)
SELECT
  logged.project_id,
  service.hours / logged.hours AS factor
FROM logged
LEFT JOIN service ON logged.project_id=service.project_id
WHERE service.hours / logged.hours < 1;

DROP TABLE IF EXISTS total_hours;
CREATE TEMPORARY TABLE total_hours AS
SELECT
  date_trunc('month', rendered_on) AS month,
  SUM(hours) AS total_hours
FROM logbook_loggedhours lh
LEFT JOIN projects_service ps ON lh.service_id=ps.id
GROUP BY month;

DROP TABLE IF EXISTS green_hours;
CREATE TEMPORARY TABLE green_hours AS
SELECT
  month,
  SUM(hours) AS green_hours
FROM (
  SELECT
    ps.project_id,
    date_trunc('month', rendered_on) AS month,
    hours * COALESCE(ghf.factor, 1) AS hours
  FROM logbook_loggedhours lh
  LEFT JOIN projects_service ps ON lh.service_id=ps.id
  LEFT JOIN projects_project p ON ps.project_id=p.id
  LEFT OUTER JOIN green_hours_factor ghf ON ghf.project_id=p.id
  WHERE
    p.type='order'
) subquery
GROUP BY month;

DROP TABLE IF EXISTS maintenance_hours;
CREATE TEMPORARY TABLE maintenance_hours AS
SELECT
  date_trunc('month', rendered_on) AS month,
  SUM(hours) AS maintenance_hours
FROM logbook_loggedhours lh
LEFT JOIN projects_service ps ON lh.service_id=ps.id
LEFT JOIN projects_project p ON ps.project_id=p.id
WHERE
  p.type='maintenance'
GROUP BY month;

DROP TABLE IF EXISTS internal_hours;
CREATE TEMPORARY TABLE internal_hours AS
SELECT
  date_trunc('month', rendered_on) AS month,
  SUM(hours) AS internal_hours
FROM logbook_loggedhours lh
LEFT JOIN projects_service ps ON lh.service_id=ps.id
LEFT JOIN projects_project p ON ps.project_id=p.id
WHERE
  p.type='internal'
GROUP BY month;

SELECT
  th.month,
  COALESCE(green_hours, 0),
  COALESCE(maintenance_hours, 0),
  COALESCE(internal_hours, 0),
  COALESCE(total_hours, 0)
FROM total_hours th
LEFT OUTER JOIN green_hours gh ON th.month=gh.month
LEFT OUTER JOIN maintenance_hours mh ON th.month=mh.month
LEFT OUTER JOIN internal_hours ih ON th.month=ih.month
ORDER BY th.month
        """
        )

        return [
            {
                "month": row[0].date(),
                "green": row[1],
                "maintenance": row[2],
                "internal": row[3],
                "total": row[4],
                "red": row[4] - row[1] - row[2] - row[3],
                "percentage": (100 * (row[1] + row[2]) / row[4]).quantize(Z0),
            }
            for row in cursor
        ]


def test():  # pragma: no cover
    import datetime as dt
    from pprint import pprint

    pprint(green_hours([dt.date(2019, 1, 1), dt.date(2019, 12, 31)]))
    pprint(green_hours_by_month())



# Generated by Django 3.1.2 on 2020-10-17 08:16

import io
import os

from django.conf import settings
from django.db import migrations


with io.open(os.path.join(settings.BASE_DIR, "stuff", "audit.sql")) as f:
    AUDIT_SQL = f.read()


class Migration(migrations.Migration):

    dependencies = [
        ("audit", "0003_auto_20190703_0813"),
    ]

    operations = [
        migrations.RunSQL(AUDIT_SQL),
    ]



import re

from django.contrib.postgres.fields import HStoreField
from django.db import models
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _


class LoggedActionQuerySet(models.QuerySet):
    def for_model(self, model):
        return self.filter(table_name=model._meta.db_table)

    def with_data(self, **kwargs):
        queryset = self
        for key, value in kwargs.items():
            queryset = queryset.filter(
                Q(**{"row_data__%s" % key: value})
                | Q(**{"changed_fields__%s" % key: value})
            )
        return queryset


class LoggedAction(models.Model):
    ACTION_TYPES = (
        ("I", "INSERT"),
        ("U", "UPDATE"),
        ("D", "DELETE"),
        ("T", "TRUNCATE"),
    )

    event_id = models.IntegerField(primary_key=True)
    table_name = models.TextField()
    user_name = models.TextField(null=True)
    created_at = models.DateTimeField()
    action = models.CharField(max_length=1, choices=ACTION_TYPES)
    row_data = HStoreField(null=True)
    changed_fields = HStoreField(null=True)

    objects = LoggedActionQuerySet.as_manager()

    class Meta:
        managed = False
        db_table = "audit_logged_actions"
        ordering = ["event_id"]
        verbose_name = _("logged action")
        verbose_name_plural = _("logged actions")

    def __str__(self):
        return "%s %s by %s at %s" % (
            self.get_action_display(),
            self.table_name,
            self.user_name,
            self.created_at,
        )

    @cached_property
    def user_id(self):
        match = re.search(r"^user-([0-9]+)-", self.user_name)
        return int(match.groups()[0]) if match else None



from workbench.tools import admin

from . import models


@admin.register(models.PlanningRequest)
class PlanningRequestAdmin(admin.ReadWriteModelAdmin):
    filter_horizontal = ["receivers"]
    raw_id_fields = ["project", "offer", "created_by"]


@admin.register(models.PlannedWork)
class PlannedWorkAdmin(admin.ReadWriteModelAdmin):
    list_display = ["project", "user", "planned_hours", "weeks"]
    raw_id_fields = ["project", "offer", "request", "user"]



# Generated by Django 3.0.1 on 2020-01-04 08:38

from django.db import migrations, models

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0019_auto_20191002_1005"),
        ("credit_control", "0006_ledger_parser"),
    ]

    operations = [
        migrations.AddField(
            model_name="creditentry",
            name="_fts",
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.RunSQL(
            search.fts(
                "credit_control_creditentry",
                ["reference_number", "total", "payment_notice", "notes", "_fts"],
            )
        ),
    ]



# Generated by Django 3.0.4 on 2020-03-13 07:24

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Note",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                ("description", models.TextField(verbose_name="description")),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.ContentType",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="notes",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created by",
                    ),
                ),
            ],
            options={
                "verbose_name": "note",
                "verbose_name_plural": "notes",
                "ordering": ["-created_at"],
            },
        ),
    ]



# Generated by Django 3.0.4 on 2020-04-02 20:12

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0017_project_cost_center"),
    ]

    operations = [
        migrations.RunSQL(search.create_structure("projects_service")),
        migrations.RunSQL(search.fts("projects_service", ["title", "description"])),
        migrations.RunSQL(
            "SELECT audit_audit_table('projects_service', ARRAY['fts_document', 'position']);"
        ),
        migrations.RunSQL("UPDATE projects_service SET id=id"),
    ]



from django.urls import re_path

from workbench import generic
from workbench.accounts.features import bookkeeping_only
from workbench.credit_control.forms import (
    AccountStatementUploadForm,
    CreditEntryForm,
    CreditEntrySearchForm,
)
from workbench.credit_control.models import CreditEntry
from workbench.credit_control.views import (
    AccountStatementUploadView,
    AssignCreditEntriesView,
)


urlpatterns = [
    re_path(
        r"^$",
        generic.ListView.as_view(
            model=CreditEntry, search_form_class=CreditEntrySearchForm
        ),
        name="credit_control_creditentry_list",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=CreditEntry),
        name="credit_control_creditentry_detail",
    ),
    re_path(
        r"^create/$",
        bookkeeping_only(
            generic.CreateView.as_view(
                model=CreditEntry,
                form_class=CreditEntryForm,
                success_url="credit_control_creditentry_list",
            )
        ),
        name="credit_control_creditentry_create",
    ),
    re_path(
        r"^upload/$",
        bookkeeping_only(
            AccountStatementUploadView.as_view(
                model=CreditEntry,
                form_class=AccountStatementUploadForm,
                success_url="credit_control_creditentry_list",
            )
        ),
        name="credit_control_creditentry_upload",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        bookkeeping_only(
            generic.UpdateView.as_view(model=CreditEntry, form_class=CreditEntryForm)
        ),
        name="credit_control_creditentry_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        bookkeeping_only(generic.DeleteView.as_view(model=CreditEntry)),
        name="credit_control_creditentry_delete",
    ),
    re_path(
        r"^assign/$",
        bookkeeping_only(AssignCreditEntriesView.as_view(model=CreditEntry)),
        name="credit_control_creditentry_assign",
    ),
]



# Generated by Django 3.1.2 on 2020-10-12 12:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0011_absence_ends_on"),
    ]

    operations = [
        migrations.AddField(
            model_name="absence",
            name="is_working_time",
            field=models.BooleanField(default=True, verbose_name="is working time"),
        ),
        migrations.AlterField(
            model_name="absence",
            name="reason",
            field=models.CharField(
                choices=[
                    ("vacation", "vacation"),
                    ("sickness", "sickness"),
                    ("paid", "paid leave (e.g. civilian service, maternity etc.)"),
                    ("other", "other reasons (no working time)"),
                    ("correction", "Working time correction"),
                ],
                max_length=10,
                verbose_name="reason",
            ),
        ),
        migrations.RunSQL(
            """
UPDATE awt_absence
SET reason='paid'
WHERE starts_on<'2020-01-01' AND reason='other';

UPDATE awt_absence
SET is_working_time=FALSE
WHERE reason='other';
            """,
            "",
        ),
    ]



# Generated by Django 2.2.4 on 2019-08-16 15:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0009_organization_is_private_person")]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="default_billing_address",
            field=models.TextField(
                blank=True,
                help_text="Mainly useful for organizations with a centralized accounts payable departement.",
                verbose_name="default billing address",
            ),
        )
    ]



import datetime as dt
from collections import defaultdict
from itertools import islice

from django.db import connections
from django.db.models import Q, Sum

from workbench.accounts.models import User
from workbench.awt.models import Absence
from workbench.invoices.utils import recurring
from workbench.logbook.models import LoggedHours
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.tools.formats import Z1, Z2, local_date_format
from workbench.tools.reporting import query
from workbench.tools.validation import monday


def _period(weeks, min, max):
    try:
        start = weeks.index(min)
    except ValueError:
        start = 0

    try:
        end = weeks.index(max)
    except ValueError:
        end = len(weeks) - 1

    return [start, end]


class Planning:
    def __init__(self, *, weeks, users=None):
        self.weeks = weeks
        self.users = users

        self._by_week = defaultdict(lambda: Z1)
        self._requested_by_week = defaultdict(lambda: Z1)
        self._by_project_and_week = defaultdict(lambda: defaultdict(lambda: Z1))
        self._projects_offers = defaultdict(lambda: defaultdict(list))
        self._project_ids = set()
        self._user_ids = {user.id for user in users} if users else set()

        self._worked_hours_by_offer = defaultdict(lambda: Z1)
        self._worked_hours_by_project = defaultdict(lambda: Z1)

        self._absences = defaultdict(lambda: [0] * len(weeks))

    def add_planned_work(self, queryset):
        for pw in queryset.filter(weeks__overlap=self.weeks).select_related(
            "user", "project__owned_by", "offer__project", "offer__owned_by", "request"
        ):
            per_week = (pw.planned_hours / len(pw.weeks)).quantize(Z2)
            for week in pw.weeks:
                self._by_week[week] += per_week
                self._by_project_and_week[pw.project][week] += per_week

            date_from = min(pw.weeks)
            date_until = max(pw.weeks) + dt.timedelta(days=6)

            self._projects_offers[pw.project][pw.offer].append(
                {
                    "work": {
                        "is_request": False,
                        "id": pw.id,
                        "request_id": pw.request_id,
                        "title": pw.title,
                        "text": pw.user.get_short_name(),
                        "user": pw.user.get_short_name(),
                        "planned_hours": pw.planned_hours,
                        "url": pw.get_absolute_url(),
                        "date_from": date_from,
                        "date_until": date_until,
                        "range": "{} – {}".format(
                            local_date_format(date_from, fmt="d.m."),
                            local_date_format(date_until, fmt="d.m."),
                        ),
                        "is_provisional": pw.request.is_provisional
                        if pw.request
                        else False,
                    },
                    "hours_per_week": [
                        per_week if week in pw.weeks else Z1 for week in self.weeks
                    ],
                    "per_week": per_week,
                }
            )

            self._project_ids.add(pw.project.pk)
            self._user_ids.add(pw.user.id)

    def add_planning_requests(self, queryset):
        for pr in (
            queryset.filter(
                Q(earliest_start_on__lte=max(self.weeks)),
                Q(completion_requested_on__gte=min(self.weeks)),
            )
            .select_related("project__owned_by", "offer__project", "offer__owned_by")
            .prefetch_related("receivers")
        ).distinct():
            per_week = (pr.missing_hours / len(pr.weeks)).quantize(Z2)
            for week in pr.weeks:
                self._requested_by_week[week] += per_week

            date_from = min(pr.weeks)
            date_until = max(pr.weeks) + dt.timedelta(days=6)

            self._projects_offers[pr.project][pr.offer].append(
                {
                    "work": {
                        "is_request": True,
                        "id": pr.id,
                        "title": pr.title,
                        "text": ", ".join(
                            user.get_short_name() for user in pr.receivers.all()
                        ),
                        "requested_hours": pr.requested_hours,
                        "planned_hours": pr.planned_hours,
                        "missing_hours": pr.missing_hours,
                        "url": pr.get_absolute_url(),
                        "date_from": date_from,
                        "date_until": date_until,
                        "range": "{} – {}".format(
                            local_date_format(date_from, fmt="d.m."),
                            local_date_format(date_until, fmt="d.m."),
                        ),
                        "period": _period(self.weeks, min(pr.weeks), max(pr.weeks)),
                        "is_provisional": pr.is_provisional,
                    },
                    "per_week": per_week,
                }
            )

            self._project_ids.add(pr.project.pk)
            self._user_ids |= {user.id for user in pr.receivers.all()}

    def add_worked_hours(self, queryset):
        for row in (
            queryset.filter(service__project__in=self._project_ids)
            .values("service__project", "service__offer")
            .annotate(Sum("hours"))
        ):
            self._worked_hours_by_offer[row["service__offer"]] += row["hours__sum"]
            self._worked_hours_by_project[row["service__project"]] += row["hours__sum"]

    def add_absences(self, queryset):
        for absence in queryset.filter(
            Q(user__in=self._user_ids),
            Q(starts_on__lte=max(self.weeks)),
            Q(ends_on__isnull=False, ends_on__gte=min(self.weeks))
            | Q(ends_on__isnull=True, starts_on__gte=min(self.weeks)),
        ).select_related("user"):
            date_from = monday(absence.starts_on)
            date_until = monday(absence.ends_on or absence.starts_on)
            hours = absence.days * absence.user.planning_hours_per_day
            weeks = [
                (idx, week)
                for idx, week in enumerate(self.weeks)
                if date_from <= week <= date_until
            ]

            for idx, week in weeks:
                self._absences[absence.user][idx] += hours / len(weeks)
                self._by_week[week] += hours / len(weeks)

    def _sort_work_list(self, work_list):
        for_requests = {}
        everything_else = []

        for row in work_list:
            if row["work"].get("request_id"):
                for_requests.setdefault(row["work"]["request_id"], []).append(row)
            else:
                everything_else.append(row)

        for row in sorted(
            everything_else,
            key=lambda row: (
                row["work"]["date_from"],
                row["work"]["date_until"],
            ),
        ):
            # Is request, has no missing hours and no planned hours in current view?
            if (
                row["work"]["is_request"]
                and not row["work"]["missing_hours"]
                and not for_requests.get(row["work"]["id"])
            ):
                continue

            yield row
            if row["work"]["is_request"] and row["work"]["id"] in for_requests:
                yield from for_requests.pop(row["work"]["id"], [])

        # Items where offers do not match or whatever
        for items in for_requests.values():
            yield from items

    def _offer_record(self, offer, work_list):
        date_from = min(pw["work"]["date_from"] for pw in work_list)
        date_until = max(pw["work"]["date_until"] for pw in work_list)
        hours = sum(
            pw["work"]["planned_hours"]
            for pw in work_list
            if not pw["work"]["is_request"]
        )

        work_list = list(self._sort_work_list(work_list))
        if not work_list:
            return None

        return {
            "offer": {
                "date_from": date_from,
                "date_until": date_until,
                "range": "{} – {}".format(
                    local_date_format(date_from, fmt="d.m."),
                    local_date_format(date_until, fmt="d.m."),
                ),
                "planned_hours": hours,
                "worked_hours": Z1,
                **(
                    {
                        "id": offer.id,
                        "title": offer.title,
                        "is_declined": offer.is_declined,
                        "is_accepted": offer.is_accepted,
                        "url": offer.get_absolute_url(),
                        "creatework": offer.project.urls["creatework"]
                        + "?offer={}".format(offer.pk),
                        "worked_hours": self._worked_hours_by_offer[offer.id],
                    }
                    if offer
                    else {}
                ),
            },
            "work_list": work_list,
        }

    def _project_record(self, project, offers):
        offers = sorted(
            filter(
                None,
                (
                    self._offer_record(offer, work_list)
                    for offer, work_list in sorted(offers.items())
                ),
            ),
            key=lambda row: (
                row["offer"]["date_from"],
                row["offer"]["date_until"],
                -row["offer"]["planned_hours"],
            ),
        )

        if not offers:
            return None

        date_from = min(rec["offer"]["date_from"] for rec in offers)
        date_until = max(rec["offer"]["date_until"] for rec in offers)
        hours = sum(rec["offer"]["planned_hours"] for rec in offers)

        return {
            "project": {
                "id": project.id,
                "title": project.title,
                "is_closed": bool(project.closed_on),
                "url": project.get_absolute_url(),
                "planning": project.urls["planning"],
                "creatework": project.urls["creatework"],
                "date_from": date_from,
                "date_until": date_until,
                "range": "{} – {}".format(
                    local_date_format(date_from, fmt="d.m."),
                    local_date_format(date_until, fmt="d.m."),
                ),
                "planned_hours": hours,
                "worked_hours": self._worked_hours_by_project[project.id],
            },
            "by_week": [
                self._by_project_and_week[project][week] for week in self.weeks
            ],
            "offers": offers,
        }

    def capacity(self):
        by_user = defaultdict(dict)
        total = defaultdict(int)

        user_ids = (
            [user.id for user in self.users] if self.users else list(self._user_ids)
        )

        for week, user, capacity in query(
            """
select week, user_id, percentage * 5 * planning_hours_per_day / 100 as capacity
from generate_series(%s::date, %s::date, '7 days') as week
left outer join lateral (
    select user_id, date_from, date_until, percentage, planning_hours_per_day
    from awt_employment
    left join accounts_user on awt_employment.user_id=accounts_user.id
    where user_id = any (%s)
) as employment
on employment.date_from <= week and employment.date_until > week
where percentage is not NULL -- NULL produced by outer join
            """,
            [min(self.weeks), max(self.weeks), user_ids],
        ):
            by_user[user][week.date()] = capacity
            total[week.date()] += capacity

        users = self.users or list(User.objects.filter(id__in=by_user))
        return {
            "total": [total.get(week, 0) for week in self.weeks],
            "by_user": [
                {
                    "user": {
                        "name": user.get_full_name(),
                        "url": user.urls["planning"],
                    },
                    "capacity": [by_user[user.id].get(week, 0) for week in self.weeks],
                }
                for user in sorted(users)
            ],
        }

    def report(self):
        try:
            this_week_index = self.weeks.index(monday())
        except ValueError:
            this_week_index = None
        return {
            "this_week_index": this_week_index,
            "weeks": [
                {
                    "month": local_date_format(week, fmt="M"),
                    "week": local_date_format(week, fmt="W"),
                    "period": "{}–{}".format(
                        local_date_format(week, fmt="j."),
                        local_date_format(week + dt.timedelta(days=6), fmt="j."),
                    ),
                }
                for week in self.weeks
            ],
            "projects_offers": sorted(
                filter(
                    None,
                    (
                        self._project_record(project, offers)
                        for project, offers in self._projects_offers.items()
                    ),
                ),
                key=lambda row: (
                    row["project"]["date_from"],
                    row["project"]["date_until"],
                    -row["project"]["planned_hours"],
                ),
            ),
            "by_week": [self._by_week[week] for week in self.weeks],
            "requested_by_week": [self._requested_by_week[week] for week in self.weeks],
            "absences": [
                (str(user), lst) for user, lst in sorted(self._absences.items())
            ],
            "capacity": self.capacity() if self.users else None,
        }


def user_planning(user):
    weeks = list(islice(recurring(monday() - dt.timedelta(days=14), "weekly"), 80))
    planning = Planning(weeks=weeks, users=[user])
    planning.add_planned_work(user.planned_work.all())
    planning.add_planning_requests(user.received_planning_requests.all())
    planning.add_worked_hours(user.loggedhours.all())
    planning.add_absences(user.absences.all())
    return planning.report()


def team_planning(team):
    weeks = list(islice(recurring(monday() - dt.timedelta(days=14), "weekly"), 80))
    planning = Planning(weeks=weeks, users=list(team.members.active()))
    planning.add_planned_work(PlannedWork.objects.filter(user__teams=team))
    planning.add_planning_requests(
        PlanningRequest.objects.filter(
            Q(receivers__teams=team) | Q(planned_work__user__teams=team)
        )
    )
    planning.add_worked_hours(LoggedHours.objects.filter(rendered_by__teams=team))
    planning.add_absences(Absence.objects.filter(user__teams=team))
    return planning.report()


def project_planning(project):
    with connections["default"].cursor() as cursor:
        cursor.execute(
            """\
WITH sq AS (
    SELECT unnest(weeks) AS week
    FROM planning_plannedwork
    WHERE project_id=%s

    UNION ALL

    SELECT earliest_start_on AS week
    FROM planning_planningrequest
    WHERE project_id=%s

    UNION ALL

    SELECT completion_requested_on - 7 AS week
    FROM planning_planningrequest
    WHERE project_id=%s
)
SELECT MIN(week), MAX(week) FROM sq
            """,
            [project.id, project.id, project.id],
        )
        result = list(cursor)[0]

    if result[0]:
        result = (min(result[0], monday() - dt.timedelta(days=14)), result[1])
        weeks = list(
            islice(
                recurring(result[0], "weekly"),
                2 + (result[1] - result[0]).days // 7,
            )
        )
    else:
        weeks = list(islice(recurring(monday() - dt.timedelta(days=14), "weekly"), 80))

    planning = Planning(weeks=weeks)
    planning.add_planned_work(project.planned_work.all())
    planning.add_planning_requests(project.planning_requests.all())
    planning.add_worked_hours(LoggedHours.objects.all())
    planning.add_absences(Absence.objects.all())
    return planning.report()


def test():  # pragma: no cover
    from pprint import pprint

    if False:
        from workbench.accounts.models import User

        pprint(user_planning(User.objects.get(pk=1)))

    if True:
        from workbench.accounts.models import Team

        pprint(team_planning(Team.objects.get(pk=1)))

    if False:
        from workbench.projects.models import Project

        pprint(project_planning(Project.objects.get(pk=8238)))



import datetime as dt

from django.core import mail
from django.core.exceptions import ValidationError
from django.test import RequestFactory, TestCase
from django.test.utils import override_settings

from workbench import factories
from workbench.contacts.models import Group, Organization, Person, PhoneNumber
from workbench.contacts.urls import autocomplete_filter
from workbench.tools.testing import messages
from workbench.tools.vcard import person_to_vcard


def person_to_dict(person, **kwargs):
    return {
        "address": person.address,
        "given_name": person.given_name,
        "family_name": person.family_name,
        "address_on_first_name_terms": person.address_on_first_name_terms,
        "salutation": person.salutation,
        "notes": person.notes,
        "organization": person.organization_id or "",
        "primary_contact": person.primary_contact_id,
        "groups": person.groups.values_list("id", flat=True),
        "is_archived": person.is_archived,
        "phonenumbers-TOTAL_FORMS": 0,
        "phonenumbers-INITIAL_FORMS": 0,
        "phonenumbers-MAX_NUM_FORMS": 1000,
        "emailaddresses-TOTAL_FORMS": 0,
        "emailaddresses-INITIAL_FORMS": 0,
        "emailaddresses-MAX_NUM_FORMS": 1000,
        "postaladdresses-TOTAL_FORMS": 0,
        "postaladdresses-INITIAL_FORMS": 0,
        "postaladdresses-MAX_NUM_FORMS": 1000,
        **kwargs,
    }


class ContactsTest(TestCase):
    def test_update(self):
        """A single-word salutation does not look right"""
        person = factories.PersonFactory.create()
        self.client.force_login(person.primary_contact)
        response = self.client.post(
            person.urls["update"], person_to_dict(person, salutation="Dear")
        )
        self.assertContains(
            response, "This does not look right. Please add a full salutation."
        )

        response = self.client.post(
            person.urls["update"], person_to_dict(person, salutation="Dear John")
        )
        self.assertRedirects(response, person.urls["detail"])

    def test_warning(self):
        """Changing the organization of a person with projects produces a warning"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create(), salutation="Dear John"
        )
        factories.ProjectFactory.create(customer=person.organization, contact=person)

        self.client.force_login(person.primary_contact)
        response = self.client.post(
            person.urls["update"], person_to_dict(person, organization="")
        )
        # print(response, response.content.decode("utf-8"))
        self.assertContains(
            response,
            "This person is the contact of the following related objects:"
            " 1 project.",
        )

        factories.Project.objects.all().delete()
        response = self.client.post(
            person.urls["update"], person_to_dict(person, organization="")
        )
        self.assertRedirects(response, person.urls["detail"])
        self.assertEqual(
            messages(response),
            ["Person 'Vorname Nachname' has been updated successfully."],
        )

    def test_lists(self):
        """The people list basically works"""
        factories.PersonFactory.create()
        person = factories.PersonFactory.create()
        group1 = Group.objects.create(title="A")
        group2 = Group.objects.create(title="B")
        person.groups.set([group1])

        person = factories.PersonFactory.create(is_archived=True)

        self.client.force_login(person.primary_contact)
        response = self.client.get(person.urls["list"])
        self.assertContains(response, "1 &ndash; 2 of 2")
        self.assertContains(response, "Vorname Nachname", 2)

        response = self.client.get("/contacts/people/?g=" + str(group1.pk))
        self.assertContains(response, person.full_name)

        response = self.client.get("/contacts/people/?g=" + str(group2.pk))
        self.assertNotContains(response, person.full_name)

        response = self.client.get("/contacts/people/?export=xlsx")
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/contacts/people/?export=vcard")
        self.assertEqual(response.status_code, 200)

    def test_organization_list(self):
        """The organization list basically works"""
        group1 = Group.objects.create(title="A")
        group2 = Group.objects.create(title="B")
        organization = factories.OrganizationFactory.create()
        organization.groups.set([group1])
        person = factories.PersonFactory.create(organization=organization)

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/contacts/organizations/")
        self.assertContains(response, str(organization))
        self.assertContains(response, person.full_name)

        response = self.client.get("/contacts/organizations/?g=" + str(group1.pk))
        self.assertContains(response, str(organization))

        response = self.client.get("/contacts/organizations/?g=" + str(group2.pk))
        self.assertNotContains(response, str(organization))

    def test_formset(self):
        """The emails formset works"""
        person = factories.PersonFactory.create()
        person.emailaddresses.create(type="E1", email="e1@example.com")
        person.emailaddresses.create(type="E2", email="e2@example.com")
        person.emailaddresses.create(type="E3", email="e3@example.com")

        self.client.force_login(person.primary_contact)
        response = self.client.get(person.urls["update"])
        # print(response, response.content.decode("utf-8"))

        self.assertContains(response, "id_emailaddresses-0-email")
        self.assertContains(response, "id_emailaddresses-1-email")
        self.assertContains(response, "id_emailaddresses-2-email")
        self.assertContains(response, "e1@example.com")
        self.assertContains(response, "e2@example.com")
        self.assertContains(response, "e3@example.com")

    def test_details(self):
        """Common methods of person details do the right thing"""
        person = factories.PersonFactory.create()
        email = person.emailaddresses.create(email="test@example.com")
        phone = person.phonenumbers.create(phone_number="012 345 678")
        address = person.postaladdresses.create(postal_address_override="A\nB")

        self.assertEqual(str(email), "test@example.com")
        self.assertEqual(str(phone), "012 345 678")
        self.assertEqual(str(address), "A\nB")

        for detail in [email, phone, address]:
            self.assertEqual(detail.get_absolute_url(), person.get_absolute_url())
            self.assertEqual(detail.urls["detail"], person.urls["detail"])

    def test_weights(self):
        """Test weighting depending on the type of person details"""
        person = factories.PersonFactory.create()
        email = person.emailaddresses.create(email="test@example.com")

        email.type = "mobil"
        email.save()
        self.assertEqual(email.weight, 30)

        email.type = "office"
        email.save()
        self.assertEqual(email.weight, 0)

        email.type = "zuhause"
        email.save()
        self.assertEqual(email.weight, 10)

        email.type = "Arbeit"
        email.save()
        self.assertEqual(email.weight, 20)

        email.type = "Firmenadresse"
        email.save()
        self.assertEqual(email.weight, -100)

    def test_contacts_redirect(self):
        """Accessing /contacts/ (not linked anywhere) redirects"""
        self.client.force_login(factories.UserFactory.create())
        self.assertRedirects(self.client.get("/contacts/"), "/contacts/people/")

    def test_phone_numbers(self):
        """Phone number validation"""
        person = factories.PersonFactory.create()
        kw = {"person": person, "type": "work"}
        with self.assertRaises(ValidationError):
            PhoneNumber(phone_number="+41 555 11 11 41 41", **kw).full_clean()
        with self.assertRaises(ValidationError):
            PhoneNumber(phone_number="+41 555 11 11 4", **kw).full_clean()
        with self.assertRaises(ValidationError):
            PhoneNumber(phone_number="just some stuff", **kw).full_clean()

        nr = PhoneNumber(phone_number="+41 555 11 11 41", **kw)
        nr.full_clean()
        self.assertEqual(nr.phone_number, "+41555111141")

        nr = PhoneNumber(phone_number="055 511 11 41", **kw)
        nr.full_clean()
        self.assertEqual(nr.phone_number, "+41555111141")

    def test_select(self):
        """The person select popup returns a redirect JSON"""
        person = factories.PersonFactory.create()
        self.client.force_login(person.primary_contact)

        response = self.client.get(person.urls["select"])
        self.assertContains(
            response, 'data-autocomplete-url="/contacts/people/autocomplete/"'
        )

        response = self.client.post(person.urls["select"], {"modal-person": person.pk})
        self.assertEqual(response.status_code, 299)
        self.assertEqual(response.json(), {"redirect": person.get_absolute_url()})

    def test_phone_number_formatting(self):
        """Phone numbers are formatted correctly, no crash on invalid data"""
        pn = PhoneNumber()
        pn.phone_number = "+41791234567"
        self.assertEqual(pn.pretty_number, "+41 79 123 45 67")
        pn.phone_number = "abcd"  # Completely bogus
        self.assertEqual(pn.pretty_number, "abcd")

    def test_person_create_redirect(self):
        """Creating persons redirects to their update page (for adding details)"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.post(
            "/contacts/people/create/",
            {
                "given_name": "Test",
                "family_name": "Jones",
                "salutation": "Hi Mr. Jones",
                "primary_contact": user.pk,
            },
        )
        self.assertEqual(response.status_code, 302)

        person = Person.objects.get()
        self.assertRedirects(response, person.urls["update"])

    def test_organization_delete_without_relations(self):
        """Organizations without relations can be deleted"""
        organization = factories.OrganizationFactory.create()
        self.client.force_login(organization.primary_contact)

        response = self.client.get(organization.urls["delete"])
        self.assertNotContains(response, "substitute_with")

        response = self.client.post(organization.urls["delete"])
        self.assertRedirects(response, organization.urls["list"])

        self.assertEqual(Organization.objects.count(), 0)

    def test_organization_delete_with_relations(self):
        """Organizations with relations can be merged with a different org."""
        organization = factories.OrganizationFactory.create()
        person = factories.PersonFactory.create(organization=organization)
        self.client.force_login(organization.primary_contact)

        group = Group.objects.create(title="A")
        organization.groups.set([group])

        response = self.client.get(organization.urls["delete"])
        self.assertContains(response, "substitute_with")

        response = self.client.post(organization.urls["delete"])
        self.assertEqual(response.status_code, 200)

        new = factories.OrganizationFactory.create()
        response = self.client.post(
            organization.urls["delete"], {"substitute_with": new.pk}
        )
        self.assertRedirects(response, organization.urls["list"])

        self.assertEqual(Organization.objects.count(), 1)

        person.refresh_from_db()
        self.assertEqual(person.organization, new)
        self.assertEqual(new.groups.count(), 0)  # Group (m2m) not assigned

    def test_create_person_with_preselected_organization(self):
        """Adding a person to an existing org. preselects the organization"""
        organization = factories.OrganizationFactory.create(name="ABCD")
        self.client.force_login(organization.primary_contact)

        response = self.client.get(
            Person.urls["create"] + "?organization={}".format(organization.pk)
        )
        self.assertContains(response, 'value="ABCD"')

    def test_organization_detail(self):
        """The organization detail page contains a few additional headings
        depending on the controlling feature"""
        organization = factories.OrganizationFactory.create()
        self.client.force_login(organization.primary_contact)

        response = self.client.get(organization.urls["detail"])
        self.assertContains(response, "Recent projects")
        self.assertContains(response, "Recent invoices")
        self.assertContains(response, "Recent offers")

        with override_settings(FEATURES={"controlling": False}):
            response = self.client.get(organization.urls["detail"])
            self.assertContains(response, "Recent projects")
            self.assertNotContains(response, "Recent invoices")
            self.assertNotContains(response, "Recent offers")

    def test_vcard_export(self):
        """Persons' details can be exported as a vCard or mailed to iOS users"""
        person = factories.PersonFactory.create()
        self.client.force_login(person.primary_contact)

        response = self.client.get(person.urls["vcard"])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "text/x-vCard;charset=utf-8")

        self.assertEqual(len(mail.outbox), 0)
        response = self.client.get(
            person.urls["vcard"],
            HTTP_REFERER=person.urls["detail"],
            HTTP_USER_AGENT="Mozilla/5.0 (iPhone; CPU iPhone OS 12_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148",  # noqa
        )
        self.assertRedirects(response, person.urls["detail"])
        self.assertEqual(len(mail.outbox), 1)

    def test_vcard(self):
        """Test aspects of the generated vCard export"""
        person = factories.PersonFactory.create(
            organization=factories.OrganizationFactory.create(),
            date_of_birth=dt.date(2020, 3, 1),
            notes="Bla",
        )

        factories.PostalAddressFactory.create(person=person)
        factories.PostalAddressFactory.create(
            person=person, postal_address_override="OVERRIDE"
        )
        person.emailaddresses.create(email="test@example.com", type="work")
        person.phonenumbers.create(phone_number="012 345 678", type="work")

        serialized = person_to_vcard(person).serialize()
        self.assertIn("OVERRIDE", serialized)
        self.assertIn("2020-03-01", serialized)

    def test_maps_url(self):
        """Person.get_maps_url returns a maps URL"""
        address = factories.PostalAddressFactory.build()
        self.assertIn("google.com/maps", address.get_maps_url())

        address = factories.PostalAddress()
        self.assertIsNone(address.get_maps_url())

    def test_person_stringification(self):
        """Persons' and postal addresses' string representation matches expectations"""

        organization = factories.OrganizationFactory.build(name="Employer A")
        person = factories.PersonFactory.build(
            organization=organization, given_name="Hans", family_name="Wurst"
        )

        self.assertEqual(str(person), "Hans Wurst")
        self.assertEqual(person.name_with_organization, "Employer A / Hans Wurst")

        home = factories.PostalAddressFactory.build(
            person=person,
            type="home",
            street="Street",
            house_number="42",
            postal_code="8000",
            city="Zürich",
        )

        work = factories.PostalAddressFactory.build(
            person=person,
            type="work",
            street="Other street",
            house_number="42",
            postal_code="8000",
            city="Zürich",
        )

        self.assertEqual(
            home.postal_address,
            """\
Hans Wurst
Street 42
8000 Zürich""",
        )

        self.assertEqual(
            work.postal_address,
            """\
Employer A
Hans Wurst
Other street 42
8000 Zürich""",
        )

        organization.is_private_person = True
        self.assertEqual(
            work.postal_address,
            """\
Hans Wurst
Other street 42
8000 Zürich""",
        )

        person = factories.PersonFactory.build(given_name="Fritz", family_name="Muster")
        self.assertEqual(str(person), "Fritz Muster")
        self.assertEqual(person.name_with_organization, "Fritz Muster")

    def test_autocomplete_filter(self):
        """autocomplete_filter test"""
        rf = RequestFactory()
        req = rf.get("/?only_employees=on")

        factories.PersonFactory.create()
        self.assertEqual(
            list(autocomplete_filter(request=req, queryset=Person.objects.all())), []
        )

    def test_archived_str(self):
        """__str__ of archived persons and organizations mentions the archival"""
        person = factories.PersonFactory.build(
            given_name="Test", family_name="Just", is_archived=True
        )
        self.assertEqual(str(person), "Test Just (archived)")

        organization = factories.OrganizationFactory.build(
            name="Test", is_archived=True
        )
        self.assertEqual(str(organization), "Test (archived)")



import datetime as dt
from decimal import Decimal

from django.core import mail
from django.core.exceptions import ValidationError
from django.test import RequestFactory, TestCase
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.planning import reporting
from workbench.planning.forms import PlannedWorkSearchForm, PlanningRequestSearchForm
from workbench.planning.models import PlannedWork, PlanningRequest
from workbench.templatetags.workbench import link_or_none
from workbench.tools.validation import monday


class PlanningTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_no_monday(self):
        """Non-mondays are rejected from PlannedWork.weeks"""
        pw = factories.PlannedWorkFactory.create(weeks=[dt.date(2020, 6, 21)])

        msg = ["Only mondays allowed, but field contains 21.06.2020."]

        with self.assertRaises(ValidationError) as cm:
            pw.clean_fields(exclude=["weeks"])
        self.assertEqual(list(cm.exception), msg)

        pw.weeks = [dt.date(2020, 6, 22)]
        pw.full_clean()  # Does not raise
        self.assertEqual(pw.pretty_from_until, "22.06.2020 – 28.06.2020")

        pr = factories.PlanningRequestFactory.create()
        pr.full_clean()

        pr.earliest_start_on = dt.date(2020, 6, 14)
        pr.completion_requested_on = dt.date(2020, 6, 21)

        with self.assertRaises(ValidationError) as cm:
            pr.clean_fields()

        msg = [
            ("earliest_start_on", ["Only mondays allowed."]),
            ("completion_requested_on", ["Only mondays allowed."]),
        ]
        self.assertEqual(list(cm.exception), msg)

        pr = factories.PlanningRequestFactory.create()
        pr.completion_requested_on = pr.earliest_start_on

        msg = ["Allow at least one week for the work please."]

        with self.assertRaises(ValidationError) as cm:
            pr.clean_fields(exclude=["completion_requested_on"])

        self.assertEqual(list(cm.exception), msg)

    def test_reporting_smoke(self):
        """Smoke test the planned work report"""
        pw = factories.PlannedWorkFactory.create(weeks=[monday()])

        factories.EmploymentFactory.create(user=pw.user, date_from=dt.date(2020, 1, 1))
        service = factories.ServiceFactory.create(project=pw.project)
        factories.LoggedHoursFactory.create(rendered_by=pw.user, service=service)
        factories.AbsenceFactory.create(user=pw.user)
        pr = factories.PlanningRequestFactory.create(
            project=pw.project,
            earliest_start_on=monday() - dt.timedelta(days=21),
            completion_requested_on=monday() + dt.timedelta(days=700),
        )
        pr.receivers.add(pw.user)

        report = reporting.user_planning(pw.user)
        self.assertAlmostEqual(sum(report["by_week"]), Decimal("26"))
        self.assertEqual(len(report["projects_offers"]), 1)

        report = reporting.project_planning(pw.project)
        self.assertAlmostEqual(sum(report["by_week"]), Decimal("26"))
        self.assertEqual(len(report["projects_offers"]), 1)

        team = factories.TeamFactory.create()
        team.members.add(pw.user)
        report = reporting.team_planning(team)
        self.assertAlmostEqual(sum(report["by_week"]), Decimal("26"))
        self.assertEqual(len(report["projects_offers"]), 1)

        pw2 = factories.PlannedWorkFactory.create(
            project=pw.project,
            user=pw.user,
            weeks=[monday()],
            request=pr,
        )
        report = reporting.user_planning(pw.user)
        self.assertAlmostEqual(sum(report["by_week"]), Decimal("46"))
        self.assertEqual(len(report["projects_offers"]), 1)

        work_list = report["projects_offers"][0]["offers"][0]["work_list"]
        self.assertEqual(len(work_list), 3)
        self.assertEqual(work_list[0]["work"]["id"], pr.id)
        self.assertEqual(work_list[1]["work"]["id"], pw2.id)
        self.assertEqual(work_list[2]["work"]["id"], pw.id)

    def test_planning_search_forms(self):
        """Planning request search form branch test"""

        project = factories.ProjectFactory.create()
        pr = factories.PlanningRequestFactory.create()
        pw = factories.PlannedWorkFactory.create(weeks=[dt.date(2020, 6, 22)])

        rf = RequestFactory()

        for form_class, model, instances in [
            (PlanningRequestSearchForm, PlanningRequest, [pr]),
            (PlannedWorkSearchForm, PlannedWork, [pw]),
        ]:
            req = rf.get("/")
            req.user = project.owned_by
            form = form_class(req.GET, request=req)
            self.assertTrue(form.is_valid())
            queryset = form.filter(model.objects.all())
            self.assertEqual(list(queryset), instances)

            req = rf.get("/?project={}".format(project.pk))
            req.user = project.owned_by
            form = form_class(req.GET, request=req)
            self.assertTrue(form.is_valid())
            queryset = form.filter(model.objects.all())
            self.assertEqual(list(queryset), [])

    def test_planning_request_form(self):
        """Create and update a planning request"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        offer = factories.OfferFactory.create()

        response = self.client.get(project.urls["createrequest"])
        self.assertNotContains(response, offer.code)
        # print(response, response.content.decode("utf-8"))

        response = self.client.post(
            project.urls["createrequest"],
            {
                "modal-title": "Request",
                "modal-earliest_start_on": "2020-06-29",
                "modal-completion_requested_on": "2020-07-27",
                "modal-requested_hours": "40",
                "modal-receivers": [offer.owned_by.id, project.owned_by.id],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        pr = PlanningRequest.objects.get()
        self.assertEqual(
            pr.weeks,
            [
                dt.date(2020, 6, 29),
                dt.date(2020, 7, 6),
                dt.date(2020, 7, 13),
                dt.date(2020, 7, 20),
            ],
        )
        self.assertEqual(set(pr.receivers.all()), {offer.owned_by, project.owned_by})

        response = self.client.post(
            pr.urls["update"],
            {
                "modal-title": "Request",
                "modal-earliest_start_on": "2020-06-29",
                "modal-completion_requested_on": "2020-07-27",
                "modal-requested_hours": "40",
                "modal-receivers": [offer.owned_by.id],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)
        self.assertEqual(set(pr.receivers.all()), {offer.owned_by})

    def test_planned_work_with_offer(self):
        """Planned work with offer preselection"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)

        response = self.client.get(
            offer.project.urls["creatework"] + "?offer={}".format(offer.pk)
        )
        self.assertContains(
            response,
            '<option value="{}" selected>{}</option>'.format(offer.pk, offer),
            html=True,
        )

    def test_planned_work_crud(self):
        """Create, update and delete planned work"""
        pr = factories.PlanningRequestFactory.create(
            earliest_start_on=monday(),
            completion_requested_on=monday() + dt.timedelta(days=14),
        )
        self.client.force_login(pr.created_by)

        response = self.client.get(pr.project.urls["creatework"] + "?request=bla")
        self.assertEqual(response.status_code, 200)  # No crash

        response = self.client.post(
            pr.project.urls["creatework"],
            {
                "modal-user": pr.created_by.id,
                "modal-title": "bla",
                "modal-planned_hours": 50,
                "modal-weeks": [monday().isoformat()],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        pw = PlannedWork.objects.get()
        response = self.client.post(
            pw.urls["update"],
            {
                "modal-user": pr.created_by.id,
                "modal-title": "bla",
                "modal-planned_hours": 50,
                "modal-weeks": [monday().isoformat()],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)

        response = self.client.post(
            pr.project.urls["creatework"] + "?request={}".format(pr.id),
            {
                "modal-request": pr.id,
                "modal-user": pr.created_by.id,
                "modal-title": "bla",
                "modal-planned_hours": 50,
                "modal-weeks": [
                    monday().isoformat(),
                    (monday() + dt.timedelta(days=7)).isoformat(),
                ],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        # print(response, response.content.decode("utf-8"))
        self.assertEqual(response.status_code, 201)

        response = self.client.post(
            pr.project.urls["creatework"] + "?request={}".format(pr.id),
            {
                "modal-request": pr.id,
                "modal-user": pr.created_by.id,
                "modal-title": "bla",
                "modal-planned_hours": 50,
                "modal-weeks": [
                    monday().isoformat(),
                    (monday() + dt.timedelta(days=14)).isoformat(),
                ],
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "weeks-outside-request")

    def test_replanning(self):
        """Moving planned work between requests"""

        pr1 = factories.PlanningRequestFactory.create(requested_hours=50)
        pr2 = factories.PlanningRequestFactory.create(requested_hours=50)

        pw = factories.PlannedWorkFactory.create(
            request=pr1, weeks=[monday()], planned_hours=20
        )

        pr1.refresh_from_db()
        pr2.refresh_from_db()
        self.assertEqual(pr1.planned_hours, 20)
        self.assertEqual(pr2.planned_hours, 0)

        pw.request = pr2
        pw.save()

        pr1.refresh_from_db()
        pr2.refresh_from_db()
        self.assertEqual(pr1.planned_hours, 0)
        self.assertEqual(pr2.planned_hours, 20)

        pw.delete()

        pr1.refresh_from_db()
        pr2.refresh_from_db()
        self.assertEqual(pr1.planned_hours, 0)
        self.assertEqual(pr2.planned_hours, 0)

    def test_receivers_with_work(self):
        """receivers_with_work returns all requested and planned work"""

        pr = factories.PlanningRequestFactory.create()
        only_receiver = factories.UserFactory.create()
        pr.receivers.add(only_receiver)
        only_pw = factories.PlannedWorkFactory.create(
            project=pr.project, request=pr, weeks=[monday()]
        )
        both = factories.PlannedWorkFactory.create(
            project=pr.project, request=pr, weeks=[monday()]
        )
        pr.receivers.add(both.user)

        self.assertEqual(set(pr.receivers.all()), {both.user, only_receiver})
        self.assertEqual(len(pr.receivers_with_work), 3)

        receivers = dict(pr.receivers_with_work)
        self.assertEqual(receivers[only_receiver], [])
        self.assertEqual(receivers[both.user], [both])
        self.assertEqual(receivers[only_pw.user], [only_pw])

    def test_planning_views(self):
        """The planning views do not crash"""
        user = factories.UserFactory.create()
        team = factories.TeamFactory.create()
        project = factories.ProjectFactory.create()

        self.client.force_login(user)

        response = self.client.get(user.urls["planning"])
        self.assertContains(response, 'id="planning-data"')

        response = self.client.get(team.urls["planning"])
        self.assertContains(response, 'id="planning-data"')

        response = self.client.get(project.urls["planning"])
        self.assertContains(response, 'id="planning-data"')

    def test_this_week_index(self):
        """this_week_index is None if current week isn't part of the report"""
        pw = factories.PlannedWorkFactory.create(weeks=[dt.date(2020, 6, 22)])

        report = reporting.project_planning(pw.project)
        self.assertIsNone(report["this_week_index"])

    def test_planning_request_notification(self):
        """Receivers of a planning request are notified"""
        pr = factories.PlanningRequestFactory.create()
        pr.receivers.add(
            factories.UserFactory.create(),
            factories.UserFactory.create(),
            factories.UserFactory.create(),
        )
        self.assertEqual(len(mail.outbox), 1)
        self.assertIn("New planning request", mail.outbox[0].subject)
        self.assertEqual(len(mail.outbox[0].to), 3)
        self.assertEqual(len(mail.outbox[0].reply_to), 4)

    def test_declined_offer_warning(self):
        """Warn when offer is declined"""
        offer = factories.OfferFactory.create(status=factories.Offer.DECLINED)
        self.client.force_login(offer.owned_by)

        response = self.client.post(
            offer.project.urls["creatework"],
            {
                "modal-user": offer.owned_by_id,
                "modal-title": "bla",
                "modal-planned_hours": 50,
                "modal-weeks": [monday().isoformat()],
                "modal-offer": offer.id,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, 'value="offer-is-declined"')

        response = self.client.post(
            offer.project.urls["createrequest"],
            {
                "modal-title": "Request",
                "modal-earliest_start_on": "2020-06-29",
                "modal-completion_requested_on": "2020-07-27",
                "modal-requested_hours": "40",
                "modal-receivers": [offer.owned_by.id],
                "modal-offer": offer.id,
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, 'value="offer-is-declined"')

    def test_link_or_none_of_planning_requests(self):
        """Planning requests have a specific implementation of link_or_none"""
        pr = factories.PlanningRequestFactory.create()
        self.assertEqual(pr.html_link(), link_or_none(pr))

    def test_planned_work_ranges(self):
        """Range detection works"""
        pw = factories.PlannedWorkFactory.create(
            weeks=[
                dt.date(2020, 6, 22),
                dt.date(2020, 6, 29),
                # dt.date(2020, 7, 6),
                dt.date(2020, 7, 13),
                dt.date(2020, 7, 20),
            ]
        )
        self.assertEqual(
            list(pw.ranges),
            [
                {
                    "from": dt.date(2020, 6, 22),
                    "until": dt.date(2020, 6, 29),
                    "pretty": "22.06.2020 – 05.07.2020",
                },
                {
                    "from": dt.date(2020, 7, 13),
                    "until": dt.date(2020, 7, 20),
                    "pretty": "13.07.2020 – 26.07.2020",
                },
            ],
        )
        self.assertEqual(pw.pretty_planned_hours, "20.0h in 4 weeks (5.0h per week)")

    def test_initialize_form_using_offer(self):
        """Initializing planning requests and planned work using offers"""

        offer = factories.OfferFactory.create(
            title="Testing title",
            description="Testing description",
        )
        self.client.force_login(offer.owned_by)

        response = self.client.get(
            offer.project.urls["createrequest"] + f"?plan_offer={offer.pk}"
        )
        self.assertContains(response, 'value="Testing title"')
        self.assertContains(response, "Testing description")

        response = self.client.get(
            offer.project.urls["creatework"] + f"?plan_offer={offer.pk}"
        )
        self.assertContains(response, 'value="Testing title"')
        self.assertContains(response, "Testing description")

        response = self.client.get(
            offer.project.urls["createrequest"] + "?plan_offer=bla"
        )
        self.assertEqual(response.status_code, 200)  # No crash

        response = self.client.get(offer.project.urls["creatework"] + "?plan_offer=bla")
        self.assertEqual(response.status_code, 200)  # No crash

    def test_initialize_form_using_service(self):
        """Initializing planning requests and planned work using services"""

        service = factories.ServiceFactory.create(
            title="Testing title",
            description="Testing description",
            effort_hours=20,
        )
        self.client.force_login(service.project.owned_by)

        response = self.client.get(
            service.project.urls["createrequest"] + f"?service={service.pk}"
        )
        self.assertContains(response, f'value="{service.project.title}: Testing title"')
        self.assertContains(response, "Testing description")
        self.assertContains(response, 'value="20.0"')

        response = self.client.get(
            service.project.urls["creatework"] + f"?service={service.pk}"
        )
        self.assertContains(response, f'value="{service.project.title}: Testing title"')
        self.assertContains(response, "Testing description")
        self.assertContains(response, 'value="20.0"')

        response = self.client.get(
            service.project.urls["createrequest"] + "?service=bla"
        )
        self.assertEqual(response.status_code, 200)  # No crash

        response = self.client.get(service.project.urls["creatework"] + "?service=bla")
        self.assertEqual(response.status_code, 200)  # No crash



from django.urls import re_path

from workbench.accounts.features import (
    FEATURES,
    controlling_only,
    deals_only,
    feature_required,
    labor_costs_only,
)
from workbench.awt.views import absence_calendar, annual_working_time_view
from workbench.circles.reporting import hours_by_circle, hours_per_work_category
from workbench.deals.reporting import accepted_deals, deal_history, declined_deals
from workbench.projects.reporting import hours_per_customer
from workbench.reporting.green_hours import green_hours
from workbench.reporting.views import (
    hours_filter_view,
    key_data_gross_profit,
    key_data_third_party_costs,
    key_data_view,
    labor_costs_view,
    logging,
    open_items_list,
    overdrawn_projects_view,
    project_budget_statistics_view,
    work_anniversaries_view,
)


urlpatterns = [
    re_path(r"^absence-calendar/$", absence_calendar, name="awt_absence_calendar"),
    re_path(
        r"^annual-working-time/$", annual_working_time_view, name="awt_year_report"
    ),
    re_path(
        r"^overdrawn-projects/$",
        controlling_only(overdrawn_projects_view),
        name="report_overdrawn_projects",
    ),
    re_path(
        r"^open-items-list/$",
        controlling_only(open_items_list),
        name="report_open_items_list",
    ),
    re_path(r"^key-data/$", controlling_only(key_data_view), name="report_key_data"),
    re_path(
        r"^key-data/gross-profit/(?P<year>[0-9]{4})\.(?P<month>[0-9]{1,2})/$",
        controlling_only(key_data_gross_profit),
    ),
    re_path(
        r"^key-data/third-party-costs/(?P<year>[0-9]{4})\.(?P<month>[0-9]{1,2})/$",
        controlling_only(key_data_third_party_costs),
    ),
    re_path(
        r"^hours-per-customer/$",
        hours_filter_view,
        {
            "template_name": "reporting/hours_per_customer.html",
            "stats_fn": hours_per_customer,
        },
        name="report_hours_per_customer",
    ),
    re_path(
        r"^hours-by-circle/$",
        feature_required(FEATURES.GLASSFROG)(hours_filter_view),
        {
            "template_name": "reporting/hours_by_circle.html",
            "stats_fn": hours_by_circle,
        },
        name="report_hours_by_circle",
    ),
    re_path(
        r"^hours-per-work-category/$",
        feature_required(FEATURES.GLASSFROG)(hours_filter_view),
        {
            "template_name": "reporting/hours_per_work_category.html",
            "stats_fn": hours_per_work_category,
        },
        name="report_hours_per_work_category",
    ),
    re_path(
        r"^project-budget-statistics/$",
        controlling_only(project_budget_statistics_view),
        name="report_project_budget_statistics",
    ),
    re_path(
        r"^green-hours/$",
        controlling_only(hours_filter_view),
        {"template_name": "reporting/green_hours.html", "stats_fn": green_hours},
        name="report_green_hours",
    ),
    re_path(
        r"^accepted-deals/$",
        deals_only(hours_filter_view),
        {"template_name": "reporting/accepted_deals.html", "stats_fn": accepted_deals},
        name="report_accepted_deals",
    ),
    re_path(
        r"^accepted-deals/deals/$",
        deals_only(hours_filter_view),
        {
            "template_name": "reporting/accepted_deals_for_user.html",
            "stats_fn": accepted_deals,
        },
    ),
    re_path(
        r"^declined-deals/$",
        deals_only(hours_filter_view),
        {"template_name": "reporting/declined_deals.html", "stats_fn": declined_deals},
        name="report_declined_deals",
    ),
    re_path(
        r"^deal-history/$",
        deals_only(hours_filter_view),
        {"template_name": "reporting/deal_history.html", "stats_fn": deal_history},
        name="report_deal_history",
    ),
    re_path(
        r"^labor-costs/$", labor_costs_only(labor_costs_view), name="report_labor_costs"
    ),
    re_path(r"^logging/$", controlling_only(logging), name="report_logging"),
    re_path(
        r"^work-anniversaries/$",
        work_anniversaries_view,
        name="report_work_anniversaries",
    ),
]



from collections import defaultdict

from django.core.management import BaseCommand
from django.db.models import Sum
from django.utils.translation import activate, gettext as _

from workbench.accounts.models import User
from workbench.contacts.models import Organization
from workbench.invoices.models import Invoice
from workbench.logbook.models import LoggedHours
from workbench.tools.formats import Z1
from workbench.tools.xlsx import WorkbenchXLSXDocument


def with_rate(row, field):
    row["rate"] = row[field] / row["hours"]
    return row


class Command(BaseCommand):
    def handle(self, **options):
        activate("de")

        invoiced_per_customer = {
            row["customer"]: row["total_excl_tax__sum"] - row["third_party_costs__sum"]
            for row in Invoice.objects.invoiced()
            .filter(project__isnull=False)
            .order_by()
            .values("customer")
            .annotate(Sum("total_excl_tax"), Sum("third_party_costs"))
        }
        hours = defaultdict(lambda: defaultdict(lambda: Z1))
        earned = defaultdict(lambda: defaultdict(lambda: Z1))
        customer_hours = defaultdict(lambda: Z1)
        user_hours = defaultdict(lambda: Z1)

        for row in (
            LoggedHours.objects.order_by()
            .values("service__project__customer", "rendered_by")
            .annotate(Sum("hours"))
        ):
            hours[row["service__project__customer"]][row["rendered_by"]] = row[
                "hours__sum"
            ]
            customer_hours[row["service__project__customer"]] += row["hours__sum"]
            user_hours[row["rendered_by"]] += row["hours__sum"]

        for customer, total_excl_tax in invoiced_per_customer.items():
            _c_hours = sum(hours[customer].values(), Z1)
            if not total_excl_tax:
                continue
            if not _c_hours:
                print(
                    "No hours for customer",
                    customer,
                    "with",
                    total_excl_tax,
                    Organization.objects.get(pk=customer),
                )
                continue
            for user, _u_hours in hours[customer].items():
                earned[customer][user] += _u_hours / _c_hours * total_excl_tax

        customers = sorted(
            (
                with_rate(
                    {
                        "customer": customer,
                        "invoiced": invoiced_per_customer.get(customer.id, Z1),
                        "hours": customer_hours[customer.id],
                    },
                    "invoiced",
                )
                for customer in Organization.objects.filter(id__in=earned.keys())
            ),
            key=lambda row: row["rate"],
            reverse=True,
        )
        users = sorted(
            (
                with_rate(
                    {
                        "user": user,
                        "earned": sum((c[user.id] for c in earned.values()), Z1),
                        "hours": user_hours[user.id],
                    },
                    "earned",
                )
                for user in User.objects.filter(id__in=user_hours)
            ),
            key=lambda row: row["rate"],
            reverse=True,
        )

        data = []
        data = [
            ["", "", "", _("user")] + [u["user"] for u in users],
            ["", "", "", _("earned")] + [u["earned"] for u in users],
            ["", "", "", _("hours")] + [u["hours"] for u in users],
            [_("customer"), _("invoiced"), _("hours"), _("rate")]
            + [u["rate"] for u in users],
        ]

        for c in customers:
            data.extend(
                [
                    [c["customer"], c["invoiced"], c["hours"], c["rate"]]
                    + [hours[c["customer"].id][u["user"].id] or None for u in users],
                    ["", "", "", ""]
                    + [earned[c["customer"].id][u["user"].id] or None for u in users],
                ]
            )

        xlsx = WorkbenchXLSXDocument()
        xlsx.add_sheet("profitability")
        xlsx.table(None, data)
        xlsx.workbook.save("profitability.xlsx")



from workbench.tools import admin

from . import models


@admin.register(models.Service)
class ServiceAdmin(admin.ModelAdmin):
    list_display = (
        "invoice",
        "project_service",
        "title",
        "description",
        "position",
        "service_hours",
        "service_cost",
    )
    raw_id_fields = ["invoice", "project_service"]


@admin.register(models.Invoice)
class InvoiceAdmin(admin.ModelAdmin):
    date_hierarchy = "invoiced_on"
    list_display = (
        "title",
        "customer",
        "invoiced_on",
        "owned_by",
        "type",
        "status",
        "total_excl_tax",
    )
    list_filter = ("type", "status")
    radio_fields = {"status": admin.HORIZONTAL, "type": admin.HORIZONTAL}
    raw_id_fields = ("customer", "contact", "project", "down_payment_applied_to")


@admin.register(models.RecurringInvoice)
class RecurringInvoiceAdmin(admin.ModelAdmin):
    list_display = [
        "title",
        "customer",
        "owned_by",
        "starts_on",
        "ends_on",
        "periodicity",
        "next_period_starts_on",
        "total_excl_tax",
    ]
    list_filter = ["periodicity"]
    radio_fields = {"periodicity": admin.HORIZONTAL}
    raw_id_fields = ["customer", "contact"]



from django.conf import settings
from django.contrib import messages
from django.db import connections
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import activate, gettext as _


def set_user_name(username):
    connections["default"].cursor().execute(
        "SET SESSION application_name TO %s", [username]
    )


def user_middleware(get_response):
    def middleware(request):
        if request.user.is_authenticated:
            set_user_name(
                "user-%d-%s" % (request.user.id, request.user.get_short_name())
            )
            activate(request.user.language)
            return get_response(request)

        set_user_name("user-0-anonymous")

        if request.path.startswith(settings.LOGIN_REQUIRED_EXEMPT):
            return get_response(request)

        messages.info(request, _("Please authenticate."))

        response = HttpResponseRedirect(reverse("login"))
        response.set_signed_cookie("next", request.get_full_path(), salt="next")
        return response

    return middleware



# Generated by Django 3.1b1 on 2020-07-11 08:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("planning", "0002_plannedwork_title"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PlanningTeamMembership",
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-22 12:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("circles", "0001_initial"),
        ("projects", "0009_service_is_optional"),
    ]

    operations = [
        migrations.AddField(
            model_name="service",
            name="role",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="services",
                to="circles.Role",
                verbose_name="role",
            ),
        )
    ]



from admin_ordering.admin import OrderableAdmin

from workbench.services.models import ServiceType
from workbench.tools import admin


@admin.register(ServiceType)
class ServiceTypeAdmin(OrderableAdmin, admin.ReadWriteModelAdmin):
    list_display = ["title", "hourly_rate", "position"]
    list_editable = ["position"]
    ordering_field = "position"



# Generated by Django 3.0.6 on 2020-05-11 11:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0018_auto_20200511_1324"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="break",
            options={
                "ordering": ["-starts_at"],
                "verbose_name": "break",
                "verbose_name_plural": "breaks",
            },
        ),
        migrations.RemoveField(
            model_name="break",
            name="ends_at",
        ),
        migrations.RemoveField(
            model_name="break",
            name="starts_at",
        ),
        migrations.RemoveField(
            model_name="break",
            name="day",
        ),
        migrations.RenameField("break", "_ends_at", "ends_at"),
        migrations.RenameField("break", "_starts_at", "starts_at"),
        migrations.AlterField(
            model_name="break",
            name="ends_at",
            field=models.DateTimeField(null=True, verbose_name="ends at"),
        ),
        migrations.AlterField(
            model_name="break",
            name="starts_at",
            field=models.DateTimeField(null=True, verbose_name="starts at"),
        ),
    ]



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import gettext as _, ngettext

from workbench import generic
from workbench.credit_control.forms import AssignCreditEntriesForm


class AccountStatementUploadView(generic.CreateView):
    def get_context_data(self, **kwargs):
        kwargs.setdefault("title", _("Upload account statement"))
        return super().get_context_data(**kwargs)

    def form_valid(self, form):
        entries = form.save()
        messages.success(
            self.request,
            ngettext(
                "Created %s credit entry.", "Created %s credit entries.", len(entries)
            )
            % len(entries),
        )
        return redirect("credit_control_creditentry_list")


class AssignCreditEntriesView(generic.CreateView):
    form_class = AssignCreditEntriesForm

    def get(self, request, *args, **kwargs):
        form = self.get_form()
        if not form.entries:
            messages.success(
                request, _("All credit entries have already been assigned.")
            )
            return redirect("credit_control_creditentry_list")
        context = self.get_context_data(form=form)
        return self.render_to_response(context)

    def form_valid(self, form):
        form.save()
        messages.success(
            self.request,
            _("%(class)s have been updated successfully.")
            % {"class": self.model._meta.verbose_name_plural},
        )
        return redirect(".")

    def get_context_data(self, **kwargs):
        kwargs.setdefault("title", _("Assign credit entries"))
        return super().get_context_data(**kwargs)



# Generated by Django 2.1.7 on 2019-03-21 13:49

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("projects", "0006_auto_20190312_1254")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("projects_project")),
        migrations.RunSQL(search.create_structure("projects_project")),
        migrations.RunSQL(search.fts("projects_project", ["title", "description"])),
    ]



import datetime as dt
from decimal import Decimal

from django.test import TestCase

from workbench import factories
from workbench.invoices.models import Invoice
from workbench.projects.models import Project
from workbench.projects.reporting import hours_per_customer, overdrawn_projects
from workbench.reporting import green_hours, project_budget_statistics
from workbench.reporting.models import Accruals
from workbench.tools.formats import Z1, Z2
from workbench.tools.testing import check_code
from workbench.tools.validation import in_days


class StatisticsTest(TestCase):
    def test_overdrawn_projects(self):
        """Overdrawn projects are overdrawn"""
        service1 = factories.ServiceFactory.create(effort_hours=2)
        factories.ServiceFactory.create(effort_hours=4)

        user = factories.UserFactory.create()

        op = list(overdrawn_projects())
        self.assertEqual(op, [])

        factories.LoggedHoursFactory.create(
            service=service1, created_by=user, rendered_by=user, hours=10
        )

        op = list(overdrawn_projects())
        self.assertEqual(len(op), 1)
        self.assertEqual(
            op,
            [
                {
                    "project": service1.project,
                    "logged_hours": Decimal("10.0"),
                    "service_hours": Decimal("2.0"),
                    "delta": Decimal("8.0"),
                }
            ],
        )

    def test_view(self):
        """Hit a few reporting views"""
        self.client.force_login(factories.UserFactory.create())

        response = self.client.get("/report/overdrawn-projects/")
        self.assertContains(response, "Overdrawn projects")

        response = self.client.get("/report/hours-per-customer/")
        self.assertContains(response, "Hours per customer")

        response = self.client.get("/report/hours-per-customer/?date_from=bla")
        self.assertRedirects(response, "/report/hours-per-customer/")

    def test_some_project_budget_statistics_view(self):
        """Project budget statistics filter form smoke test"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        factories.ProjectFactory.create()
        factories.ProjectFactory.create(closed_on=dt.date.today())

        response = self.client.get("/report/project-budget-statistics/")
        self.assertContains(response, "Project budget statistics")

        code = check_code(self, "/report/project-budget-statistics/")
        code("owned_by=-1")
        code("owned_by=0")
        code("owned_by={}".format(user.pk))
        code("owned_by=bla", 302)

        code("closed_during_the_last_year=on")
        code("internal=on")
        code("export=xlsx")

    def test_not_archived_hours_grouped_services_green_hours_hpc(self):
        """Test a scenario"""
        service1 = factories.ServiceFactory.create(effort_rate=180, effort_type="Any")
        service2 = factories.ServiceFactory.create(project=service1.project)

        project = Project.objects.get()
        self.assertEqual(
            project.not_archived_total, {"total": Z2, "hours_rate_undefined": Z1}
        )

        factories.LoggedHoursFactory.create(service=service1, hours=10)
        factories.LoggedHoursFactory.create(service=service2, hours=20)
        factories.LoggedCostFactory.create(
            service=service1, cost=0, third_party_costs=0
        )

        project = Project.objects.get()
        self.assertEqual(
            project.not_archived_total,
            {"total": Decimal("1800.00"), "hours_rate_undefined": Decimal("20.00")},
        )

        invoice = factories.InvoiceFactory.create(
            project=project,
            customer=project.customer,
            contact=project.contact,
            type=factories.Invoice.SERVICES,
            invoiced_on=in_days(0),
            status=Invoice.SENT,
        )
        invoice.create_services_from_logbook(project.services.all())

        project = Project.objects.get()
        self.assertEqual(
            project.not_archived_total, {"total": Z2, "hours_rate_undefined": Z1}
        )

        hours = factories.LoggedHoursFactory.create(service=service1, hours=10)
        project = Project.objects.get()
        self.assertEqual(
            project.not_archived_total,
            {"total": Decimal("1800.00"), "hours_rate_undefined": Z1},
        )

        self.assertEqual(project.project_invoices_total_excl_tax, Decimal("1800.00"))

        grouped = project.grouped_services
        self.assertEqual(len(grouped["offers"]), 1)
        self.assertIs(grouped["offers"][0][0], None)  # Offer is None
        self.assertEqual(grouped["logged_hours"], Decimal(40))
        self.assertEqual(grouped["service_hours"], 0)
        self.assertEqual(grouped["total_logged_cost"], Decimal(3600))
        self.assertEqual(grouped["total_service_cost"], 0)
        self.assertEqual(grouped["total_logged_hours_rate_undefined"], Decimal(20))
        self.assertEqual(grouped["total_service_hours_rate_undefined"], 0)

        today = dt.date.today()
        date_range = [dt.date(today.year, 1, 1), dt.date(today.year, 12, 31)]

        hpc = hours_per_customer(date_range)
        self.assertEqual(hpc["organizations"][0]["total_hours"], Decimal(40))
        self.assertEqual(len(hpc["organizations"]), 1)
        self.assertEqual(len(hpc["users"]), 3)

        hpc = hours_per_customer(date_range, users=[hours.rendered_by])
        self.assertEqual(hpc["organizations"][0]["total_hours"], Decimal(10))
        self.assertEqual(len(hpc["organizations"]), 1)
        self.assertEqual(len(hpc["users"]), 1)

        stats = project_budget_statistics.project_budget_statistics(
            Project.objects.all()
        )
        self.assertEqual(
            stats["statistics"],
            [
                {
                    "cost": Decimal("0.00"),
                    "delta": Decimal("1800.000"),
                    "effort_cost": Decimal("3600.000"),
                    "effort_hours_with_rate_undefined": Decimal("20.00"),
                    "hours": Decimal("40.00"),
                    "invoiced": Decimal("1800.00"),
                    "logbook": Decimal("3600.000"),
                    "not_archived": Decimal("10.0"),
                    "offered": Decimal("0.00"),
                    "project": project,
                    "third_party_costs": Decimal("0.00"),
                }
            ],
        )
        self.assertEqual(stats["overall"]["delta_positive"], Decimal("1800.00"))
        self.assertEqual(stats["overall"]["delta_negative"], Decimal("0.00"))

    def create_projects(self):
        p_internal = factories.ProjectFactory.create(type=Project.INTERNAL)
        p_maintenance = factories.ProjectFactory.create(type=Project.MAINTENANCE)
        p_order = factories.ProjectFactory.create(type=Project.ORDER)

        s_internal = factories.ServiceFactory.create(project=p_internal)
        s_maintenance = factories.ServiceFactory.create(project=p_maintenance)
        s_order = factories.ServiceFactory.create(project=p_order, effort_hours=20)

        factories.LoggedHoursFactory.create(
            service=s_internal, hours=10, rendered_on=dt.date(2019, 1, 1)
        )
        factories.LoggedHoursFactory.create(
            service=s_maintenance, hours=20, rendered_on=dt.date(2019, 1, 1)
        )
        factories.LoggedHoursFactory.create(
            service=s_order, hours=5, rendered_on=dt.date(2019, 1, 1)
        )
        factories.LoggedHoursFactory.create(
            service=s_order, hours=25, rendered_on=dt.date(2019, 2, 1)
        )
        factories.LoggedHoursFactory.create(
            service=s_order, hours=10, rendered_on=dt.date(2019, 3, 1)
        )

        factories.LoggedHoursFactory.create(
            service=s_order, hours=10, rendered_on=dt.date(2018, 12, 1)
        )
        factories.LoggedHoursFactory.create(
            service=s_order, hours=10, rendered_on=dt.date(2019, 4, 1)
        )

    def test_green_hours(self):
        """Green hours report incl. filtering and overall stats"""
        self.create_projects()
        p_green = factories.ProjectFactory.create(type=Project.ORDER)
        s_green = factories.ServiceFactory.create(project=p_green, effort_hours=20)
        factories.LoggedHoursFactory.create(
            service=s_green, hours=10, rendered_on=dt.date(2019, 1, 1)
        )

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/report/green-hours/")
        self.assertEqual(response.status_code, 200)
        response = self.client.get(
            "/report/green-hours/?team={}".format(factories.TeamFactory.create().pk)
        )
        self.assertEqual(response.status_code, 200)

        gh = green_hours.green_hours([dt.date(2019, 1, 1), dt.date(2019, 3, 31)])
        overall = gh[-1]
        self.assertEqual(overall[0], 0)
        self.assertEqual(overall[1]["total"], Decimal("80"))
        self.assertEqual(overall[1]["maintenance"], Decimal("20"))
        self.assertAlmostEqual(overall[1]["green"], Decimal("23.33333333"))
        self.assertAlmostEqual(overall[1]["red"], Decimal("26.66666666"))
        self.assertAlmostEqual(overall[1]["percentage"], Decimal("54.16666666"))

        gh = green_hours.green_hours_by_month()
        self.assertEqual(len(gh), 5)
        self.assertAlmostEqual(gh[-1]["green"], Decimal(10) / 3)
        self.assertAlmostEqual(gh[-1]["red"], Decimal(10) * 2 / 3)
        self.assertEqual(gh[-1]["internal"], Decimal(0))
        self.assertEqual(gh[-1]["maintenance"], Decimal(0))
        self.assertEqual(gh[-1]["total"], Decimal(10))
        self.assertAlmostEqual(gh[-1]["percentage"], Decimal(33))

        self.client.force_login(p_green.owned_by)
        # Also run the FTE code branch
        p_green.owned_by.employments.create(
            percentage=50, vacation_weeks=5, date_from=dt.date(2010, 1, 1)
        )

        # More branches
        invoice = factories.InvoiceFactory.create(
            subtotal=100, third_party_costs=10, status=factories.Invoice.SENT
        )
        Accruals.objects.for_cutoff_date(dt.date.today())
        factories.LoggedCostFactory.create(description="ASDF", third_party_costs=1)

        self.assertEqual(self.client.get("/report/key-data/").status_code, 200)

        month = dt.date.today().strftime("%Y.%m")
        response = self.client.get("/report/key-data/gross-profit/%s/" % month)
        self.assertContains(response, invoice.title)

        response = self.client.get("/report/key-data/third-party-costs/%s/" % month)
        self.assertContains(response, invoice.title)
        self.assertContains(response, "ASDF")
        # print(response, response.content.decode("utf-8"))

    def test_statistics(self):
        """The project statistics modal does not crash"""
        hours = factories.LoggedHoursFactory.create()
        self.client.force_login(hours.rendered_by)
        response = self.client.get(hours.service.project.urls["statistics"])
        self.assertEqual(response.status_code, 200)



# Generated by Django 2.1.7 on 2019-03-12 08:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0003_auto_20190305_0809")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="is_archived",
            field=models.BooleanField(
                default=False,
                verbose_name="is archived",
                help_text="It is preferrable to archive records instead of deleting them.",
            ),
        )
    ]



# Generated by Django 3.1.2 on 2020-10-21 07:22

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("timer", "0007_auto_20200506_2130"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="timestamp",
            name="project",
        ),
    ]



import datetime as dt
from decimal import Decimal

from django.core import mail
from django.test import RequestFactory, TestCase

from time_machine import travel

from workbench import factories
from workbench.reporting.accounting import send_accounting_files
from workbench.reporting.labor_costs import labor_costs_by_cost_center
from workbench.reporting.models import Accruals
from workbench.reporting.views import DateRangeAndTeamFilterForm


class ReportingTest(TestCase):
    def test_send_accounting_files(self):
        """Accounting files are automatically sent on the first day of the year"""
        factories.UserFactory.create(is_admin=True)
        factories.ProjectFactory.create()

        with travel("2019-12-26 12:00"):
            send_accounting_files()
            self.assertEqual(len(mail.outbox), 0)

        with travel("2020-01-01 12:00"):
            send_accounting_files()
            self.assertEqual(len(mail.outbox), 1)

    def test_accruals(self):
        """Accrual calculation does not crash..."""
        obj = Accruals.objects.for_cutoff_date(dt.date.today())
        self.assertEqual(obj.accruals, Decimal("0.00"))
        self.assertEqual(str(obj), obj.cutoff_date.strftime("%d.%m.%Y"))

    def test_labor_costs(self):
        """The labor costs report does a few things"""
        user1 = factories.EmploymentFactory.create().user
        user2 = factories.EmploymentFactory.create(
            hourly_labor_costs=100, green_hours_target=75
        ).user

        service = factories.ServiceFactory.create()

        factories.LoggedHoursFactory.create(service=service, rendered_by=user1)
        factories.LoggedHoursFactory.create(service=service, rendered_by=user2)
        factories.LoggedCostFactory.create(
            service=service, third_party_costs=10, cost=15
        )

        self.client.force_login(user1)
        response = self.client.get("/report/labor-costs/")
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            "/report/labor-costs/?project={}".format(service.project_id)
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get(
            "/report/labor-costs/?cost_center={}".format(
                factories.CostCenterFactory.create().pk
            )
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.get("/report/labor-costs/?users=all")
        self.assertEqual(response.status_code, 200)

        year = dt.date.today().year
        lc = labor_costs_by_cost_center([dt.date(year, 1, 1), dt.date(year, 12, 31)])
        lcp = lc["cost_centers"][0]["projects"]

        self.assertEqual(len(lcp), 1)
        self.assertAlmostEqual(lcp[0]["costs"], Decimal("100"))
        self.assertAlmostEqual(
            lcp[0]["costs_with_green_hours_target"], Decimal("133.3333333")
        )
        self.assertEqual(lcp[0]["hours"], 2)
        self.assertEqual(lcp[0]["hours_with_rate_undefined"], 1)

        self.assertEqual(lcp[0]["third_party_costs"], Decimal("10"))
        self.assertEqual(lcp[0]["revenue"], Decimal("115"))

    def test_teams_filter(self):
        """Filtering by teams and individuals"""
        rf = RequestFactory()
        req = rf.get("/")

        user1 = factories.UserFactory.create()
        user2 = factories.UserFactory.create()
        team = factories.TeamFactory.create()
        team.members.add(user1)

        form = DateRangeAndTeamFilterForm({"team": team.id}, request=req)
        self.assertTrue(form.is_valid())
        self.assertEqual(set(form.users()), {user1})

        form = DateRangeAndTeamFilterForm({"team": -user2.id}, request=req)
        self.assertTrue(form.is_valid())
        self.assertEqual(set(form.users()), {user2})



import datetime as dt
from collections import defaultdict

from django.contrib import messages
from django.shortcuts import redirect, render
from django.utils.translation import gettext, ngettext
from django.views.decorators.http import require_POST

from workbench import generic
from workbench.invoices.models import Invoice
from workbench.logbook.models import LoggedCost, LoggedHours
from workbench.tools.pdf import pdf_response
from workbench.tools.xlsx import WorkbenchXLSXDocument


class InvoicePDFView(generic.DetailView):
    model = Invoice

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        pdf, response = pdf_response(
            self.object.code,
            as_attachment=request.GET.get("disposition") == "attachment",
        )

        pdf.init_letter()
        pdf.process_invoice(self.object)
        pdf.generate()

        return response


class InvoiceXLSXView(generic.DetailView):
    model = Invoice

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        data = [
            [
                gettext("service"),
                gettext("description"),
                gettext("rendered on"),
                gettext("effort type"),
                gettext("hourly rate"),
                gettext("hours"),
                gettext("cost"),
            ],
            [],
        ]

        hours = defaultdict(list)
        cost = defaultdict(list)

        for entry in LoggedHours.objects.filter(invoice_service__invoice=self.object):
            hours[entry.invoice_service_id].append(entry)
        for entry in LoggedCost.objects.filter(invoice_service__invoice=self.object):
            cost[entry.invoice_service_id].append(entry)

        for service in self.object.services.all():
            data.append(
                [
                    service.title,
                    service.description,
                    "",
                    service.effort_type,
                    service.effort_rate,
                    "",
                    "",
                ]
            )
            for entry in hours[service.id]:
                data.append(
                    [
                        "",
                        entry.description,
                        entry.rendered_on,
                        "",
                        "",
                        entry.hours,
                        entry.hours * service.effort_rate
                        if service.effort_rate
                        else "",
                    ]
                )
            for entry in cost[service.id]:
                data.append(
                    ["", entry.description, entry.rendered_on, "", "", "", entry.cost]
                )
            data.append([])

        xlsx = WorkbenchXLSXDocument()
        xlsx.add_sheet(gettext("logbook"))
        xlsx.table(None, data)
        return xlsx.to_response("%s.xlsx" % self.object.code)


class RecurringInvoiceDetailView(generic.DetailView):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        if request.GET.get("create_invoices"):
            invoices = self.object.create_invoices()
            messages.info(
                request,
                ngettext("Created %s invoice.", "Created %s invoices.", len(invoices))
                % len(invoices),
            )
            return redirect("invoices_invoice_list" if len(invoices) else self.object)
        context = self.get_context_data()
        return self.render_to_response(context)


def reminders(request):
    invoices = Invoice.objects.overdue().select_related(
        "customer", "owned_by", "project"
    )
    by_organization = {}
    for invoice in invoices:
        if invoice.customer not in by_organization:
            by_organization[invoice.customer] = {
                "organization": invoice.customer,
                "last_reminded_on": {invoice.last_reminded_on},
                "invoices": [invoice],
            }
        else:
            row = by_organization[invoice.customer]
            row["invoices"].append(invoice)
            row["last_reminded_on"].add(invoice.last_reminded_on)

    def last_reminded_on(row):
        days = row["last_reminded_on"] - {None}
        return max(days) if days else None

    reminders = [
        dict(row, last_reminded_on=last_reminded_on(row))
        for row in by_organization.values()
    ]

    return render(
        request,
        "invoices/reminders.html",
        {
            "reminders": sorted(
                reminders,
                key=lambda row: (
                    row["last_reminded_on"] or dt.date.min,
                    row["organization"].name,
                ),
            )
        },
    )


@require_POST
def dunning_letter(request, customer_id):
    invoices = (
        Invoice.objects.overdue()
        .filter(customer=customer_id)
        .select_related("customer", "contact__organization", "owned_by", "project")
    )

    pdf, response = pdf_response("reminders", as_attachment=True)
    pdf.dunning_letter(invoices=list(invoices))
    pdf.generate()

    invoices.update(last_reminded_on=dt.date.today())

    return response



# Generated by Django 3.0.2 on 2020-01-22 11:59

import django.db.models
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="TimerState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "state",
                    django.db.models.JSONField(default=dict, verbose_name="state"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="updated at"),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "timer state",
                "verbose_name_plural": "timer states",
            },
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-15 10:58

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("logbook", "0014_auto_20190915_1250")]

    operations = [migrations.RemoveField(model_name="loggedcost", name="project")]



from collections import defaultdict
from decimal import Decimal

from django.db.models import Sum
from django.utils.translation import gettext as _

from workbench.accounts.models import User
from workbench.circles.models import Circle, Role
from workbench.logbook.models import LoggedHours
from workbench.tools.formats import Z1


def hours_by_circle(date_range, *, users=None):
    queryset = LoggedHours.objects.order_by().filter(rendered_on__range=date_range)
    if users:
        queryset = queryset.filter(rendered_by__in=users)

    queryset = queryset.values("service__role", "rendered_by").annotate(Sum("hours"))
    seen_users = set()

    hours_by_role = defaultdict(lambda: defaultdict(Decimal))
    hours_by_circle = defaultdict(lambda: defaultdict(Decimal))

    roles_to_circle = {role.id: role.circle_id for role in Role.objects.all()}

    for row in queryset:
        hours_by_role[row["service__role"]][row["rendered_by"]] = row["hours__sum"]
        hours_by_circle[roles_to_circle.get(row["service__role"])][
            row["rendered_by"]
        ] += row["hours__sum"]
        seen_users.add(row["rendered_by"])

    users = list(User.objects.filter(id__in=seen_users))

    circles = [
        {
            "users": users,
            "circle": None,
            "by_user": [(user, hours_by_circle[None][user.id]) for user in users],
            "total": sum(hours_by_circle[None].values(), Decimal()),
        }
    ]
    for circle in Circle.objects.prefetch_related("roles"):
        row = {
            "circle": circle,
            "roles": [],
            "by_user": [(user, hours_by_circle[circle.id][user.id]) for user in users],
            "total": sum(hours_by_circle[circle.id].values(), Decimal()),
        }

        for role in circle.roles.all():
            role_total = sum(hours_by_role[role.id].values(), Decimal())
            if role_total:
                row["roles"].append(
                    {
                        "role": role,
                        "by_user": [
                            (user, hours_by_role[role.id][user.id]) for user in users
                        ],
                        "total": role_total,
                    }
                )
        if row["total"]:
            circles.append(row)
    return {
        "circles": circles,
        "total_hours": sum((circle["total"] for circle in circles), Decimal()),
    }


def hours_per_work_category(date_range, *, users=None):
    queryset = LoggedHours.objects.order_by().filter(rendered_on__range=date_range)
    if users:
        queryset = queryset.filter(rendered_by__in=users)

    queryset = queryset.values("service__role__work_category", "rendered_by").annotate(
        Sum("hours")
    )
    seen_users = set()

    by_user_and_category = defaultdict(lambda: defaultdict(Decimal))
    by_category = defaultdict(Decimal)

    for row in queryset:
        by_user_and_category[row["rendered_by"]][
            row["service__role__work_category"] or None
        ] = row["hours__sum"]
        by_category[row["service__role__work_category"] or None] += row["hours__sum"]
        seen_users.add(row["rendered_by"])

    users = [
        (
            user,
            {
                "total": sum(by_user_and_category[user.id].values(), Z1),
                "per_category": [
                    (row[0], by_user_and_category[user.id][row[0]])
                    for row in Role.CATEGORIES
                ],
                "undefined": by_user_and_category[user.id][None],
            },
        )
        for user in User.objects.filter(id__in=seen_users)
    ]
    return {
        "categories": Role.CATEGORIES,
        "users": users,
        "chart": [
            {
                "label": title,
                "hours": [
                    100 * by_user_and_category[user.id][name] / user_stats["total"]
                    for user, user_stats in users
                ],
            }
            for name, title, description in Role.CATEGORIES
        ]
        + [
            {
                "label": _("Undefined"),
                "hours": [
                    100 * by_user_and_category[user.id][None] / user_stats["total"]
                    for user, user_stats in users
                ],
            }
        ],
        "summary": {
            "total": sum(by_category.values(), Z1),
            "per_category": [(row[0], by_category[row[0]]) for row in Role.CATEGORIES],
            "undefined": by_category[None],
        },
    }



from django.test import TestCase
from django.test.utils import override_settings

from workbench import factories
from workbench.tools.search import process_query


class SearchTest(TestCase):
    def test_search(self):
        """Searching works and only shows allowed content objects"""
        project = factories.ProjectFactory.create(title="Test")
        self.client.force_login(project.owned_by)

        response = self.client.get("/search/")
        self.assertContains(response, "Search query missing.")

        response = self.client.get("/search/?q=Test")
        self.assertContains(response, project.get_absolute_url())

        self.assertContains(response, "projects")
        self.assertContains(response, "organizations")
        self.assertContains(response, "people")
        self.assertContains(response, "invoices")
        self.assertContains(response, "recurring-invoices")
        self.assertContains(response, "offers")
        self.assertContains(response, "deals")

        with override_settings(FEATURES={"controlling": False}):
            response = self.client.get("/search/?q=Test")
            self.assertContains(response, project.get_absolute_url())

            self.assertContains(response, "projects")
            self.assertContains(response, "organizations")
            self.assertContains(response, "people")
            self.assertNotContains(response, "invoices")
            self.assertNotContains(response, "recurring-invoices")
            self.assertNotContains(response, "offers")
            self.assertNotContains(response, "deals")

    def test_process_query(self):
        """Specific testing of query->tsquery conversion"""
        self.assertEqual(process_query(""), "")
        self.assertEqual(process_query("org"), "org:*")
        self.assertEqual(process_query("a b"), "a & b:*")
        self.assertEqual(process_query("(foo bar)"), "foo & bar:*")



import datetime as dt
import io
import zipfile

from django.http import HttpResponse
from django.utils.formats import date_format
from django.utils.text import slugify
from django.utils.translation import gettext as _

from workbench.tools.formats import days, hours, local_date_format
from workbench.tools.pdf import PDFDocument, mm
from workbench.tools.xlsx import WorkbenchXLSXDocument


def annual_working_time_pdf(statistics):
    if len(statistics["statistics"]) == 1:
        response = HttpResponse(
            user_stats_pdf(statistics["statistics"][0]), content_type="application/pdf"
        )
        response["Content-Disposition"] = 'attachment; filename="awt.pdf"'
        return response

    with io.BytesIO() as buf:
        with zipfile.ZipFile(buf, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            for data in statistics["statistics"]:
                zf.writestr(
                    "%s-%s.pdf"
                    % (
                        slugify(data["user"].get_full_name()),
                        data["months"]["year"].year,
                    ),
                    user_stats_pdf(data),
                )
            xlsx = WorkbenchXLSXDocument()
            xlsx.add_sheet(_("running net work hours"))
            xlsx.table(
                [""]
                + [date_format(day, "M") for day in data["months"]["months"]]
                + [_("vacation days credit")],
                [
                    [data["user"].get_full_name()]
                    + data["running_sums"]
                    + [data["totals"]["vacation_days_credit"]]
                    for data in statistics["statistics"]
                ],
            )
            with io.BytesIO() as x:
                xlsx.workbook.save(x)
                zf.writestr("statistics.xlsx", x.getvalue())
        response = HttpResponse(buf.getvalue(), content_type="application/zip")
        response["Content-Disposition"] = 'attachment; filename="awt.zip"'
        return response


def user_stats_pdf(data):
    with io.BytesIO() as buf:
        pdf = PDFDocument(buf, font_size=7)
        pdf.init_report()

        awt_columns = [9.5 * mm for i in range(12)]
        awt_columns.append(12 * mm)
        awt_columns.insert(0, pdf.bounds.E - pdf.bounds.W - sum(awt_columns))
        awt_table_style = pdf.style.tableHead + (("TOPPADDING", (0, 1), (-1, -1), 3),)

        pdf.h1(data["user"])
        pdf.spacer(1 * mm)
        pdf.p(_("annual working time"))
        pdf.spacer(3 * mm)

        table = []
        table.append(
            [data["months"]["year"]]
            + [date_format(day, "M") for day in data["months"]["months"]]
            + [_("Total")]
        )

        table.append(
            [_("target days for full time employment")]
            + data["months"]["target_days"]
            + [data["totals"]["target_days"]]
        )
        table.append(
            [_("pensum")]
            + ["%.0f%%" % p for p in data["months"]["percentage"]]
            + ["%.0f%%" % data["totals"]["percentage"]]
        )
        table.append(
            [_("vacation days available")]
            + [days(value) for value in data["months"]["available_vacation_days"]]
            + [days(data["totals"]["available_vacation_days"])]
        )
        table.append(
            [
                "%s (%s)"
                % (
                    _("target time"),
                    data["months"]["year"].pretty_working_time_per_day,
                )
            ]
            + [hours(value) for value in data["months"]["target"]]
            + [hours(data["totals"]["target"])]
        )
        table.append(
            [_("vacation days taken")]
            + [
                days(value) if value else ""
                for value in data["months"]["absence_vacation"]
            ]
            + [days(data["totals"]["absence_vacation"])]
        )
        table.append(
            [_("sickness days")]
            + [
                days(value) if value else ""
                for value in data["months"]["absence_sickness"]
            ]
            + [days(data["totals"]["absence_sickness"])]
        )
        if data["totals"]["absence_paid"]:
            table.append(
                [_("Paid leave")]
                + [
                    days(value) if value else ""
                    for value in data["months"]["absence_paid"]
                ]
                + [days(data["totals"]["absence_paid"])]
            )
        if data["totals"]["absence_correction"]:
            table.append(
                [_("Working time correction")]
                + [
                    days(value) if value else ""
                    for value in data["months"]["absence_correction"]
                ]
                + [days(data["totals"]["absence_correction"])]
            )
        if data["totals"]["vacation_days_correction"]:
            table.append(
                [_("vacation days correction")]
                + [
                    days(value) if value else ""
                    for value in data["months"]["vacation_days_correction"]
                ]
                + [days(data["totals"]["vacation_days_correction"])]
            )
        table.append(
            [_("countable absence hours")]
            + [hours(value) if value else "" for value in data["absences_time"]]
            + [hours(data["totals"]["absences_time"])]
        )
        table.append(
            [_("logged hours")]
            + [hours(value) if value else "" for value in data["months"]["hours"]]
            + [hours(data["totals"]["hours"])]
        )
        table.append(
            [_("working time")]
            + [hours(value) if value else "" for value in data["working_time"]]
            + [hours(data["totals"]["working_time"])]
        )
        table.append(
            [_("net work hours per month")]
            + [hours(value) for value in data["monthly_sums"]]
            + [""]
        )

        pdf.table(table, awt_columns, awt_table_style)
        pdf.spacer(1 * mm)

        table = []
        table.append(
            [_("running net work hours")]
            + [hours(value) for value in data["running_sums"]]
            + [hours(data["totals"]["running_sum"])]
        )

        if data["totals"]["vacation_days_credit"]:
            table.append(
                [_("vacation days credit")]
                + [""] * 12
                + [days(data["totals"]["vacation_days_credit"])]
            )

        table.append([_("balance")] + [""] * 12 + [hours(data["totals"]["balance"])])

        pdf.table(
            table,
            awt_columns,
            awt_table_style + (("FONT", (0, 0), (-1, 0), "Rep", pdf.style.fontSize),),
        )
        pdf.spacer()

        table = []
        for employment in data["employments"]:
            table.append(
                [
                    employment,
                    employment.percentage,
                    "%.0f" % employment.vacation_weeks,
                    employment.notes,
                ]
            )

        if table:
            pdf.table(
                [[_("employment"), _("percentage"), _("vacation weeks"), _("notes")]]
                + table,
                pdf.table_columns((35 * mm, 15 * mm, 20 * mm, None)),
                awt_table_style + (("ALIGN", (0, 0), (-1, -1), "LEFT"),),
            )
            pdf.spacer()

        for key, reason in [
            ("absence_vacation", _("vacation days")),
            ("absence_sickness", _("sickness days")),
            ("absence_paid", _("Paid leave")),
            ("absence_correction", _("Working time correction")),
        ]:
            table = [
                [absence.pretty_period, days(absence.days), absence.description]
                for absence in data["absences"][key]
            ]
            if table:
                pdf.table(
                    [[reason, "", ""]] + table,
                    pdf.table_columns((35 * mm, 15 * mm, None)),
                    awt_table_style + (("ALIGN", (0, 0), (-1, -1), "LEFT"),),
                )
                pdf.spacer()

        pdf.p(_("Generated on %(day)s") % {"day": local_date_format(dt.date.today())})

        pdf.generate()
        return buf.getvalue()



# Generated by Django 2.1.7 on 2019-03-21 13:47

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("credit_control", "0002_auto_20190308_1752")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("credit_control_creditentry")),
        migrations.RunSQL(search.create_structure("credit_control_creditentry")),
        migrations.RunSQL(
            search.fts(
                "credit_control_creditentry",
                ["reference_number", "total", "payment_notice", "notes"],
            )
        ),
    ]



# Copied 1:1 from django-user-payments

import datetime as dt
import itertools


def next_valid_day(year, month, day):
    """
    Return the next valid date for the given year, month and day combination.

    Used by ``recurring`` below.
    """
    while True:
        try:
            return dt.date(year, month, day)
        except ValueError:
            if month > 12:
                month -= 12
                year += 1
                continue

            day += 1
            if day > 31:
                day = 1
                month += 1


def recurring(start, periodicity):
    """
    This generator yields valid dates with the given start date and
    periodicity.

    Returns later dates if calculated dates do not exist, e.g. for a yearly
    periodicity and a starting date of 2016-02-29 (a leap year), dates for
    years that aren't leap years will be 20xx-03-01, not 20xx-02-28. However,
    leap year dates will stay on 20xx-02-29 and not be delayed.
    """
    if periodicity == "yearly":
        return (  # pragma: no branch
            next_valid_day(start.year + i, start.month, start.day)
            for i in itertools.count()
        )

    elif periodicity == "quarterly":
        return (  # pragma: no branch
            next_valid_day(start.year, start.month + i * 3, start.day)
            for i in itertools.count()
        )

    elif periodicity == "monthly":
        return (  # pragma: no branch
            next_valid_day(start.year, start.month + i, start.day)
            for i in itertools.count()
        )

    elif periodicity == "weekly":
        return (  # pragma: no branch
            start + dt.timedelta(days=i * 7) for i in itertools.count()
        )

    else:
        raise ValueError("Unknown periodicity %r" % periodicity)



# Generated by Django 2.1.7 on 2019-03-06 19:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("projects", "0002_auto_20190304_2247")]

    operations = [
        migrations.AlterModelOptions(
            name="cost",
            options={
                "ordering": ["pk"],
                "verbose_name": "cost",
                "verbose_name_plural": "costs",
            },
        ),
        migrations.AlterModelOptions(
            name="effort",
            options={
                "ordering": ["pk"],
                "verbose_name": "effort",
                "verbose_name_plural": "efforts",
            },
        ),
        migrations.RemoveField(model_name="cost", name="position"),
        migrations.AlterUniqueTogether(name="effort", unique_together=set()),
    ]



# Generated by Django 2.2.2 on 2019-06-30 20:13

from django.db import migrations, models

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("projects", "0011_auto_20190629_0943")]

    operations = [
        migrations.AddField(
            model_name="project",
            name="_fts",
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.RunSQL(
            search.fts("projects_project", ["title", "description", "_fts"])
        ),
    ]



from django.urls import re_path

from workbench import generic
from workbench.accounts.features import controlling_only
from workbench.invoices.forms import RecurringInvoiceForm, RecurringInvoiceSearchForm
from workbench.invoices.models import RecurringInvoice
from workbench.invoices.views import RecurringInvoiceDetailView


urlpatterns = [
    re_path(
        r"^$",
        controlling_only(
            generic.ListView.as_view(
                model=RecurringInvoice, search_form_class=RecurringInvoiceSearchForm
            )
        ),
        name="invoices_recurringinvoice_list",
    ),
    re_path(
        r"^(?P<pk>\d+)/$",
        controlling_only(RecurringInvoiceDetailView.as_view(model=RecurringInvoice)),
        name="invoices_recurringinvoice_detail",
    ),
    re_path(
        r"^create/$",
        controlling_only(
            generic.CreateView.as_view(
                model=RecurringInvoice, form_class=RecurringInvoiceForm
            )
        ),
        name="invoices_recurringinvoice_create",
    ),
    re_path(
        r"^(?P<pk>\d+)/update/$",
        controlling_only(
            generic.UpdateView.as_view(
                model=RecurringInvoice, form_class=RecurringInvoiceForm
            )
        ),
        name="invoices_recurringinvoice_update",
    ),
    re_path(
        r"^(?P<pk>\d+)/delete/$",
        controlling_only(generic.DeleteView.as_view(model=RecurringInvoice)),
        name="invoices_recurringinvoice_delete",
    ),
]



from workbench.contacts.models import (
    EmailAddress,
    Group,
    Organization,
    Person,
    PhoneNumber,
    PostalAddress,
)
from workbench.tools import admin


class PhoneNumberInline(admin.TabularInline):
    model = PhoneNumber
    extra = 0


class EmailAddressInline(admin.TabularInline):
    model = EmailAddress
    extra = 0


class PostalAddressInline(admin.TabularInline):
    model = PostalAddress
    extra = 0


@admin.register(Organization)
class OrganizationAdmin(admin.ModelAdmin):
    filter_horizontal = ("groups",)
    list_display = ("name", "primary_contact")
    raw_id_fields = ("primary_contact",)
    search_fields = ("name", "notes")


@admin.register(Person)
class PersonAdmin(admin.ModelAdmin):
    filter_horizontal = ("groups",)
    inlines = [PhoneNumberInline, EmailAddressInline, PostalAddressInline]
    list_display = ["given_name", "family_name", "organization", "primary_contact"]
    list_filter = ["is_archived"]
    raw_id_fields = ["organization", "primary_contact"]
    search_fields = ["given_name", "family_name", "notes"]


@admin.register(Group)
class GroupAdmin(admin.ReadWriteModelAdmin):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-24 07:20
from __future__ import unicode_literals

import io
import os

import django.contrib.postgres.fields.hstore
from django.conf import settings
from django.contrib.postgres.operations import HStoreExtension, UnaccentExtension
from django.db import migrations, models


with io.open(os.path.join(settings.BASE_DIR, "stuff", "audit.sql")) as f:
    AUDIT_SQL = f.read()


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        HStoreExtension(),
        UnaccentExtension(),
        migrations.CreateModel(
            name="LoggedAction",
            fields=[
                ("event_id", models.IntegerField(primary_key=True, serialize=False)),
                ("table_name", models.TextField()),
                ("user_name", models.TextField(null=True)),
                ("created_at", models.DateTimeField()),
                (
                    "action",
                    models.CharField(
                        choices=[
                            ("I", "INSERT"),
                            ("U", "UPDATE"),
                            ("D", "DELETE"),
                            ("T", "TRUNCATE"),
                        ],
                        max_length=1,
                    ),
                ),
                (
                    "row_data",
                    django.contrib.postgres.fields.hstore.HStoreField(null=True),
                ),
                (
                    "changed_fields",
                    django.contrib.postgres.fields.hstore.HStoreField(null=True),
                ),
            ],
            options={
                "verbose_name_plural": "logged actions",
                "ordering": ["event_id"],
                "verbose_name": "logged action",
                "managed": False,
                "db_table": "audit_logged_actions",
            },
        ),
        migrations.RunSQL(AUDIT_SQL),
    ]



import datetime as dt
from decimal import Decimal

from django.utils.formats import date_format
from django.utils.timezone import localtime
from django.utils.translation import ngettext


Z0 = Decimal("0")
Z1 = Decimal("0.0")
Z2 = Decimal("0.00")


def local_date_format(dttm, *, fmt=None):
    if not dttm:
        return ""
    if hasattr(dttm, "astimezone"):
        dttm = localtime(dttm)
    return date_format(
        dttm, fmt or ("d.m.Y H:i" if isinstance(dttm, dt.datetime) else "d.m.Y")
    )


def _fmt(*, fmt, value, exp, plus_sign=False):
    value = value.quantize(exp) if value else exp
    value = value if value != 0 else exp  # Avoid -0.0
    return fmt.format("+" if plus_sign and value > 0 else "", value).replace(",", "’")


def currency(value, plus_sign=False):
    return _fmt(fmt="{}{:,.2f}", value=value, exp=Z2, plus_sign=plus_sign)


def days(value, plus_sign=False):
    return _fmt(fmt="{}{:,.2f}d", value=value, exp=Z2, plus_sign=plus_sign)


def hours(value, plus_sign=False):
    return _fmt(fmt="{}{:,.1f}h", value=value, exp=Z1, plus_sign=plus_sign)


def hours_and_minutes(seconds):
    minutes = int(seconds) // 60
    hours, minutes = divmod(minutes, 60)
    parts = [
        (ngettext("%s hour", "%s hours", hours) % hours) if hours else None,
        ngettext("%s minute", "%s minutes", minutes) % minutes,
    ]
    return " ".join(filter(None, parts))



import datetime as dt

from django.contrib import messages
from django.db import models
from django.utils.translation import gettext, gettext_lazy as _

from workbench.accounts.features import FEATURES
from workbench.accounts.models import User
from workbench.tools.formats import hours, local_date_format
from workbench.tools.models import HoursField, Model, MoneyField
from workbench.tools.urls import model_urls
from workbench.tools.validation import raise_if_errors


class WorkingTimeModel(models.Model):
    name = models.CharField(_("name"), max_length=100)
    position = models.IntegerField(_("position"), default=0)

    class Meta:
        ordering = ["position", "id"]
        verbose_name = _("working time model")
        verbose_name_plural = _("working time models")

    def __str__(self):
        return self.name


class Year(Model):
    MONTHS = [
        "january",
        "february",
        "march",
        "april",
        "may",
        "june",
        "july",
        "august",
        "september",
        "october",
        "november",
        "december",
    ]

    working_time_model = models.ForeignKey(
        WorkingTimeModel,
        on_delete=models.CASCADE,
        verbose_name=_("working time model"),
    )
    year = models.IntegerField(_("year"))
    january = models.DecimalField(_("january"), max_digits=4, decimal_places=2)
    february = models.DecimalField(_("february"), max_digits=4, decimal_places=2)
    march = models.DecimalField(_("march"), max_digits=4, decimal_places=2)
    april = models.DecimalField(_("april"), max_digits=4, decimal_places=2)
    may = models.DecimalField(_("may"), max_digits=4, decimal_places=2)
    june = models.DecimalField(_("june"), max_digits=4, decimal_places=2)
    july = models.DecimalField(_("july"), max_digits=4, decimal_places=2)
    august = models.DecimalField(_("august"), max_digits=4, decimal_places=2)
    september = models.DecimalField(_("september"), max_digits=4, decimal_places=2)
    october = models.DecimalField(_("october"), max_digits=4, decimal_places=2)
    november = models.DecimalField(_("november"), max_digits=4, decimal_places=2)
    december = models.DecimalField(_("december"), max_digits=4, decimal_places=2)
    working_time_per_day = HoursField(_("working time per day"))

    class Meta:
        ordering = ["-year"]
        unique_together = (("working_time_model", "year"),)
        verbose_name = _("year")
        verbose_name_plural = _("years")

    def __str__(self):
        return "%s %s" % (self.year, self.working_time_model)

    @property
    def months(self):
        return [getattr(self, field) for field in self.MONTHS]

    @property
    def pretty_working_time_per_day(self):
        return "%s/%s" % (hours(self.working_time_per_day), _("day"))


class Employment(Model):
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        verbose_name=_("user"),
        related_name="employments",
    )
    date_from = models.DateField(_("date from"), default=dt.date.today)
    date_until = models.DateField(_("date until"), default=dt.date.max)
    percentage = models.IntegerField(_("percentage"))
    vacation_weeks = models.DecimalField(
        _("vacation weeks"),
        max_digits=4,
        decimal_places=2,
        help_text=_("Vacation weeks for a full year."),
    )
    notes = models.CharField(_("notes"), blank=True, max_length=500)
    hourly_labor_costs = MoneyField(_("hourly labor costs"), blank=True, null=True)
    green_hours_target = models.SmallIntegerField(
        _("green hours target"), blank=True, null=True
    )

    class Meta:
        ordering = ["date_from"]
        unique_together = ["user", "date_from"]
        verbose_name = _("employment")
        verbose_name_plural = _("employments")

    def __str__(self):
        if self.date_until.year > 3000:
            return _("since %s") % local_date_format(self.date_from)
        return "%s - %s" % (
            local_date_format(self.date_from),
            local_date_format(self.date_until),
        )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        next = None
        for employment in self.user.employments.reverse():
            if next is None:
                next = employment
            else:
                if employment.date_until >= next.date_from:
                    employment.date_until = next.date_from - dt.timedelta(days=1)
                    super(Employment, employment).save()
                next = employment

    save.alters_data = True

    def clean_fields(self, exclude):
        super().clean_fields(exclude)
        errors = {}
        if (self.hourly_labor_costs is None) != (self.green_hours_target is None):
            errors["__all__"] = _(
                "Either provide both hourly labor costs"
                " and green hours target or none."
            )
        if self.date_from and self.date_until and self.date_from > self.date_until:
            errors["date_until"] = _("Employments cannot end before they began.")
        raise_if_errors(errors, exclude)


@model_urls
class Absence(Model):
    VACATION = "vacation"
    SICKNESS = "sickness"
    PAID = "paid"
    OTHER = "other"
    CORRECTION = "correction"

    REASON_CHOICES = [
        (VACATION, _("vacation")),
        (SICKNESS, _("sickness")),
        (PAID, _("paid leave (e.g. civilian service, maternity etc.)")),
        (OTHER, _("other reasons (no working time)")),
        (CORRECTION, _("Working time correction")),
    ]

    user = models.ForeignKey(
        User, on_delete=models.CASCADE, verbose_name=_("user"), related_name="absences"
    )
    starts_on = models.DateField(_("starts on"))
    ends_on = models.DateField(_("ends on"), blank=True, null=True)
    days = models.DecimalField(_("days"), max_digits=4, decimal_places=2)
    description = models.TextField(_("description"))
    reason = models.CharField(_("reason"), max_length=10, choices=REASON_CHOICES)
    is_vacation = models.BooleanField(_("is vacation"), default=True)
    is_working_time = models.BooleanField(_("is working time"), default=True)

    class Meta:
        ordering = ["-starts_on", "-pk"]
        verbose_name = _("absence")
        verbose_name_plural = _("absences")

    def __str__(self):
        return self.description

    def save(self, *args, **kwargs):
        self.is_vacation = self.reason == self.VACATION
        self.is_working_time = self.reason != self.OTHER
        super().save(*args, **kwargs)

    save.alters_data = True

    @classmethod
    def allow_update(cls, instance, request):
        if (
            instance.user.enforce_same_week_logging
            and instance.starts_on.year < dt.date.today().year
        ):
            messages.error(request, _("Absences of past years are locked."))
            return False
        if (
            instance.reason == instance.CORRECTION
            and not request.user.features[FEATURES.WORKING_TIME_CORRECTION]
        ):
            messages.error(
                request,
                _(
                    "You are not permitted to edit absences"
                    " of type «Working time correction»."
                ),
            )
            return False
        return True

    allow_delete = allow_update

    @classmethod
    def get_redirect_url(cls, instance, request):
        if not request.is_ajax():
            return cls.urls["list"] + "?u={}".format(
                instance.user_id if instance else ""
            )

    def clean_fields(self, exclude):
        super().clean_fields(exclude)
        errors = {}
        if self.starts_on and self.ends_on:
            if self.ends_on < self.starts_on:
                errors["ends_on"] = _("Absences cannot end before they began.")
            if self.starts_on.year != self.ends_on.year:
                errors["ends_on"] = _("Start and end must be in the same year.")
        raise_if_errors(errors, exclude)

    @property
    def pretty_status(self):
        return "%s, %s" % (gettext("%s days") % self.days, self.pretty_period)

    @property
    def pretty_period(self):
        return "%s - %s" % (
            local_date_format(self.starts_on),
            local_date_format(self.ends_on or self.starts_on),
        )



# Generated by Django 2.2.2 on 2019-07-09 19:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("expenses", "0001_initial"),
        ("logbook", "0009_auto_20190628_1659"),
    ]

    operations = [
        migrations.RemoveField(model_name="loggedcost", name="expenses_reimbursed_at"),
        migrations.AddField(
            model_name="loggedcost",
            name="expense_report",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="expenses",
                to="expenses.ExpenseReport",
                verbose_name="expense report",
            ),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-11 12:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0019_auto_20200511_1330"),
    ]

    operations = [
        migrations.AlterField(
            model_name="break",
            name="ends_at",
            field=models.DateTimeField(verbose_name="ends at"),
        ),
        migrations.AlterField(
            model_name="break",
            name="starts_at",
            field=models.DateTimeField(verbose_name="starts at"),
        ),
    ]



from django.conf import settings

from workbench.accounts.features import FEATURES


def workbench(request):
    return {
        "WORKBENCH": settings.WORKBENCH,
        "FEATURES": FEATURES,
        "DEBUG": settings.DEBUG,
        "TESTING": settings.TESTING,
    }



# Generated by Django 2.2rc1 on 2019-03-27 08:45

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("awt", "0003_auto_20190312_1254")]

    operations = [
        migrations.RunSQL(search.create_structure("awt_absence")),
        migrations.RunSQL(search.fts("awt_absence", ["description"])),
        migrations.RunSQL("UPDATE awt_absence SET id=id"),
    ]



# Generated by Django 2.1.7 on 2019-03-04 21:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("projects", "0001_initial"), ("audit", "0001_initial")]

    operations = [
        migrations.RunSQL(
            "SELECT audit_audit_table('projects_project');"
            "SELECT audit_audit_table('projects_service');"
            "SELECT audit_audit_table('projects_effort');"
            "SELECT audit_audit_table('projects_cost');",
            "",
        )
    ]



# Generated by Django 2.1.7 on 2019-03-21 13:45

from django.db import migrations

from workbench.tools import search


class Migration(migrations.Migration):

    dependencies = [("contacts", "0005_auto_20190321_1025")]

    operations = [
        migrations.RunSQL(search.drop_old_shit("contacts_organization")),
        migrations.RunSQL(search.drop_old_shit("contacts_person")),
        migrations.RunSQL(search.create_structure("contacts_organization")),
        migrations.RunSQL(search.create_structure("contacts_person")),
        migrations.RunSQL(search.fts("contacts_organization", ["name"])),
        migrations.RunSQL(
            search.fts(
                "contacts_person", ["given_name", "family_name", "address", "notes"]
            )
        ),
    ]



# Generated by Django 3.0 on 2019-12-07 13:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("awt", "0007_auto_20191207_1307"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="workingtimemodel",
            options={
                "ordering": ["position", "id"],
                "verbose_name": "working time model",
                "verbose_name_plural": "working time models",
            },
        ),
        migrations.AddField(
            model_name="workingtimemodel",
            name="position",
            field=models.IntegerField(default=0, verbose_name="position"),
        ),
    ]



import operator
from collections import OrderedDict
from decimal import Decimal
from functools import reduce
from itertools import count

from django import forms
from django.conf import settings
from django.contrib import messages
from django.http import JsonResponse
from django.shortcuts import redirect
from django.utils.text import capfirst
from django.utils.translation import gettext as _

from workbench import generic
from workbench.expenses.models import ExchangeRates, ExpenseReport
from workbench.tools.formats import Z2, currency, local_date_format
from workbench.tools.pdf import MarkupParagraph, mm, pdf_response


class ExpenseReportPDFView(generic.DetailView):
    model = ExpenseReport

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if not self.object.closed_on:
            messages.warning(
                request,
                _(
                    "Please close the expense report first. Generating PDFs"
                    " for open expense reports isn't allowed."
                ),
            )
            return redirect(self.object)

        pdf, response = pdf_response(
            self.object.code,
            as_attachment=request.GET.get("disposition") == "attachment",
        )
        pdf.init_report()
        pdf.h1(_("expense report"))
        pdf.spacer(2 * mm)
        pdf.table(
            [
                (capfirst(_("of")), self.object.owned_by.get_full_name()),
                (capfirst(_("created at")), local_date_format(self.object.created_at)),
                (capfirst(_("status")), capfirst(self.object.pretty_status)),
            ],
            pdf.style.tableColumnsLeft,
            pdf.style.table,
        )
        pdf.spacer(5 * mm)

        counter = count(1)
        expenses = OrderedDict()
        for cost in self.object.expenses.select_related(
            "service__project__owned_by"
        ).order_by("rendered_on", "pk"):
            expenses.setdefault(cost.expense_currency, []).append(cost)

        for currency_code, sublist in sorted(expenses.items()):
            pdf.table(
                [(_("receipt"), "", "")]
                + [
                    (
                        "%d." % (next(counter),),
                        MarkupParagraph(
                            "%s<br />%s: %s<br />%s<br />&nbsp;"
                            % (
                                local_date_format(cost.rendered_on),
                                cost.service.project,
                                cost.service,
                                cost.description,
                            ),
                            pdf.style.normal,
                        ),
                        currency(cost.third_party_costs)
                        if cost.expense_cost is None
                        else currency(cost.expense_cost),
                    )
                    for cost in sublist
                ],
                (10 * mm, pdf.bounds.E - pdf.bounds.W - 10 * mm - 16 * mm, 16 * mm),
                pdf.style.tableHead,
            )

            pdf.spacer(0.7 * mm)
            total_cost = reduce(
                operator.add,
                (cost.expense_cost or cost.third_party_costs for cost in sublist),
                Z2,
            )
            pdf.table(
                [
                    (
                        "%s %s"
                        % (
                            capfirst(_("total")),
                            currency_code or settings.WORKBENCH.CURRENCY,
                        ),
                        currency(total_cost),
                    )
                ],
                pdf.style.tableColumns,
                pdf.style.tableHeadLine,
            )
            pdf.spacer()

        pdf.generate()

        return response


class ConvertForm(forms.Form):
    day = forms.DateField()
    currency = forms.CharField()
    cost = forms.DecimalField(max_digits=10, decimal_places=2)


def convert(request):
    form = ConvertForm(request.GET)
    if not form.is_valid():
        return JsonResponse({"cost": ""}, status=400)
    rates = ExchangeRates.objects.for_day(form.cleaned_data["day"])
    cost = form.cleaned_data["cost"] / Decimal(
        str(rates.rates["rates"][form.cleaned_data["currency"]])
    )
    return JsonResponse({"cost": cost.quantize(Z2)})



from contextlib import suppress

from django.urls import NoReverseMatch, reverse


class _MUHelper:
    def __init__(self, viewname_pattern, kwargs):
        self.viewname_pattern = viewname_pattern
        self.kwargs = kwargs

    def __getitem__(self, item):
        viewname = self.viewname_pattern.format(item)
        if self.kwargs:
            with suppress(NoReverseMatch):
                return reverse(viewname, **self.kwargs)
        return reverse(viewname)


class _Descriptor:
    def __init__(self, viewname_pattern):
        self.viewname_pattern = viewname_pattern

    def __get__(self, obj, objtype=None):
        if not obj:
            return _MUHelper(self.viewname_pattern, {})
        kwargs = {"kwargs": {"pk": obj.pk}}
        helper = obj.__dict__["urls"] = _MUHelper(self.viewname_pattern, kwargs)
        return helper


def model_urls(cls):
    """
    Usage::

        @model_urls
        class MyModel(models.Model):
            pass

        instance = MyModel.objects.get(...)
        instance.urls["detail"] == instance.get_absolute_url()
    """

    cls.urls = _Descriptor(
        viewname_pattern="{}_{}_{{}}".format(cls._meta.app_label, cls._meta.model_name)
    )
    if not hasattr(cls, "get_absolute_url"):
        cls.get_absolute_url = lambda self: self.urls["detail"]
    return cls



# Generated by Django 2.2.3 on 2019-07-21 08:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("accounts", "0003_user_enforce_same_week_logging")]

    operations = [
        migrations.AddField(
            model_name="user",
            name="language",
            field=models.CharField(
                choices=[("en", "english"), ("de", "german")],
                max_length=10,
                verbose_name="language",
            ),
        )
    ]



# Generated by Django 3.1b1 on 2020-07-10 15:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contacts", "0012_person_date_of_birth"),
    ]

    operations = [
        migrations.AddField(
            model_name="organization",
            name="is_archived",
            field=models.BooleanField(
                default=False,
                help_text="It is preferrable to archive records instead of deleting them.",
                verbose_name="is archived",
            ),
        ),
    ]



from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from workbench.tools import admin

from . import models


def content_object_url(instance):
    try:
        url = instance.get_absolute_url()
    except Exception:
        return instance.content_type.name
    else:
        return format_html('<a href="{}">{}</a>', url, instance.content_type.name)


content_object_url.short_description = _("content object")


@admin.register(models.Note)
class NoteAdmin(admin.ModelAdmin):
    list_display = ["title", content_object_url, "created_by", "created_at"]
    list_select_related = ["content_type", "created_by"]
    search_fields = ["title", "description"]



# Generated by Django 2.1.7 on 2019-03-12 11:54

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("invoices", "0007_auto_20190311_1456")]

    operations = [
        migrations.AlterField(
            model_name="service",
            name="effort_hours",
            field=models.DecimalField(
                blank=True,
                decimal_places=1,
                max_digits=10,
                null=True,
                verbose_name="hours",
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="service_hours",
            field=models.DecimalField(
                decimal_places=1, default=0, max_digits=10, verbose_name="service hours"
            ),
        ),
    ]



# Generated by Django 2.2 on 2019-06-28 14:59

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("logbook", "0008_auto_20190518_1443")]

    operations = [
        migrations.AlterField(
            model_name="loggedhours",
            name="created_at",
            field=models.DateTimeField(
                db_index=True,
                default=django.utils.timezone.now,
                verbose_name="created at",
            ),
        )
    ]



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.tools.models import Model, SearchQuerySet
from workbench.tools.urls import model_urls


class NoteQuerySet(SearchQuerySet):
    def for_content_object(self, content_object):
        return self.filter(
            content_type=ContentType.objects.get_for_model(content_object),
            object_id=content_object.pk,
        )


@model_urls
class Note(Model):
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    created_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        verbose_name=_("created by"),
        related_name="notes",
    )
    title = models.CharField(_("title"), max_length=200)
    description = models.TextField(_("description"))

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    objects = NoteQuerySet.as_manager()

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("note")
        verbose_name_plural = _("notes")

    def __str__(self):
        return "{} - {} {} - {}".format(
            self.title,
            self.content_type.name,
            self.object_id,
            self.created_by.get_short_name(),
        )

    def get_absolute_url(self):
        model = self.content_type.model_class()
        viewname = "%s_%s_detail" % (model._meta.app_label, model._meta.model_name)
        return reverse(viewname, kwargs={"pk": self.object_id})



import datetime as dt
from decimal import Decimal

from django.test import TestCase

from workbench import factories
from workbench.logbook.reporting import classify_logging_delay, logbook_stats
from workbench.tools.formats import Z1


class ReportingTest(TestCase):
    def test_report(self):
        """Smoke test the logging stats report"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        factories.LoggedHoursFactory.create()

        response = self.client.get("/report/logging/")
        self.assertContains(response, "Logging statistics")

    def test_stats(self):
        """_Very_ basic testing of logged hours statistics"""
        stats = logbook_stats([dt.date.today(), dt.date.today()])

        self.assertEqual(len(stats["users"]), 0)
        self.assertEqual(
            stats["logged_hours_stats"], {"count": 0, "sum": Z1, "avg": None}
        )

        factories.LoggedHoursFactory.create()
        factories.LoggedHoursFactory.create()

        stats = logbook_stats([dt.date.today(), dt.date.today()])

        self.assertEqual(len(stats["users"]), 2)
        self.assertEqual(
            stats["users"][0]["logged_hours_stats"],
            {"avg": Decimal("1"), "count": 1, "sum": Decimal("1.0")},
        )

    def test_classify_logging_delay(self):
        """Logging delay classification"""
        self.assertEqual(classify_logging_delay(Decimal(-1))[1], "success")
        self.assertEqual(classify_logging_delay(Decimal(4))[1], "light")
        self.assertEqual(classify_logging_delay(Decimal(10))[1], "caveat")
        self.assertEqual(classify_logging_delay(Decimal(40))[1], "danger")
